/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <stdio.h>
#include <sys/ioctl.h>
#include <fcntl.h>
#include <signal.h>
#include <errno.h>
#include <string.h>
#include <strings.h>

#include "listener.h"
#include "version.h"

#include <exLayout.h>
#include <exErrors.h>
#include <exFkm.h>
#include <exMenu.h>
#include <exEdit.h>

#include <Xm/DrawingA.h>
#include <Xm/Form.h>
#include <Xm/PanedW.h>

#include <X11/Shell.h>

static char SccsID[] = "%W%\t%G%";            /* sccs keyword */

int displayIcon();

static void listenerTitle();
static void initialise();
static void insertMenuReturnValue();
static void insertFunctionKeyValue();
static void writePrompt();

static GC normal_gc;

BufInfo *bufinfoptr();

char msg[128];

/*initCLdescriptions: ********************************************************
 * initialise the editor description structures used to create the editors 
 * of  the command listener 
 *****************************************************************************/

initCLdescriptions(lview, lcommand)
     EditorDescr *lview, *lcommand;
{
  lview->inputFnct = (int (*)()) NULL; 
  lview->name = Malloc(strlen(ESH_NAME) + 2 + strlen(" (output window)"));
  sprintf(lview->name, "%s%s", ESH_NAME, " (output window)");
  lview->file = (char *) 0;
  lview->buffer = (char **) 0;
  lview->bufferLines = 500;
  lview->lines = 24;
  lview->cols = 70;
  lview->flags = (ED_MSGWINDOW | ED_SCROLLBAR | ED_HELP | ED_LINEBREAK | ED_AUTOKILL);

  lcommand->inputFnct = (int (*)()) NULL;
  lcommand->name = Malloc(strlen(ESH_NAME) + 2 + strlen(" (command editor)"));
  sprintf(lcommand->name, "%s%s", ESH_NAME, " (command editor)");
  lcommand->file = (char *) 0;
  lcommand->buffer = (char **) 0;
  lcommand->bufferLines = 20;
  lcommand->lines = 5;
  lcommand->cols = lview->cols - PROMPTWIDTH;
  lcommand->flags = lview->flags;
}

/*initMHfunctions: ************************************************************
 * initialise the functions array that allows the menu system to access 
 * local functions 
 *****************************************************************************/

initMHfunctions()
{
  int cnt = 0;

  /* esh - menu functions */
  MH_FUNCTIONS[cnt].func = (int (*)()) 0;
  MH_FUNCTIONS[cnt].funcName = (char *) 0; cnt++;
  MH_FUNCTIONS[cnt].func = loadApplNode;
  MH_FUNCTIONS[cnt].funcName = "loadApplNode"; cnt++;
  MH_FUNCTIONS[cnt].func = loadMacroMenu;
  MH_FUNCTIONS[cnt].funcName = "loadMacroMenu"; cnt++;
  MH_FUNCTIONS[cnt].func = hideApplNode;
  MH_FUNCTIONS[cnt].funcName = "hideApplNode"; cnt++;
  MH_FUNCTIONS[cnt].func = hideMacroMenu;
  MH_FUNCTIONS[cnt].funcName = "hideMacroMenu"; cnt++;
  MH_FUNCTIONS[cnt].func = loadFunctionKeySet;
  MH_FUNCTIONS[cnt].funcName = "loadFunctionKeySet"; cnt++;
  MH_FUNCTIONS[cnt].func = closeFunctionKeySet;
  MH_FUNCTIONS[cnt].funcName = "closeFunctionKeySet"; cnt++;
  MH_FUNCTIONS[cnt].func = newApplication;
  MH_FUNCTIONS[cnt].funcName = "newApplication"; cnt++;
  MH_FUNCTIONS[cnt].func = loadApplication;
  MH_FUNCTIONS[cnt].funcName = "loadApplication"; cnt++;
  MH_FUNCTIONS[cnt].func = quitApplication;
  MH_FUNCTIONS[cnt].funcName = "quitApplication"; cnt++;
  MH_FUNCTIONS[cnt].func = eshQuit;
  MH_FUNCTIONS[cnt].funcName = "eshQuit"; cnt++;
  MH_FUNCTIONS[cnt].func = startShell;
  MH_FUNCTIONS[cnt].funcName = "startShell"; cnt++;
  MH_FUNCTIONS[cnt].func = (int (*)()) 0;
  MH_FUNCTIONS[cnt].funcName = (char *) 0; cnt++;
}

/* createListener: ***********************************************************
 * create the listener structure 
 ****************************************************************************/

Listener *createListener(parent, view, command, geometry)
     Widget parent;
     EditorDescr *view,			/* view component of listener */
       *command;			/* command entry of listener */
     char *geometry;			/* user geometry specification */
{
  char fullName[512];
  Listener *l;
  int cnt, h;
  int x = 0, y = 0;
  int cols = 0, rows = 0;
  Arg args[20];
  Widget w;
  XmFontList fl = NULL;  
  BufInfo *bf_view, *bf_command;
  static XtTranslations ptt = NULL;

  static char prompt_translations[] = 
    " <Expose> 	: writePrompt()";

  static XtActionsRec prompt_actions[] = {
    { "writePrompt", writePrompt}
  };

  if (! ptt) {
    XtAddActions(prompt_actions, 1);
    ptt = XtParseTranslationTable(prompt_translations);
  }

  initialise(XtDisplay(parent));
  
  /* get geometry specification from command line */
  XParseGeometry(geometry, &x, &y, &cols, &rows);

  l = (Listener *) Malloc(sizeof(Listener));
  
  l->popupCnt = MDS_MAXPOPUP;
  l->parent = parent;
  l->display = XtDisplay(parent);
  l->dpyno = XConnectionNumber(l->display);
  l->gc = normal_gc;

  /* create application shell */
  cnt = 0;
  XtSetArg(args[cnt], XmNx, x); cnt++;
  XtSetArg(args[cnt], XmNy, y); cnt++;
  XtSetArg(args[cnt], XmNiconName, VERSION); cnt++;
  XtSetArg(args[cnt], XmNiconX, 1060); cnt++;
  XtSetArg(args[cnt], XmNiconY, 770); cnt++;
  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
  l->listener_shell = XtCreatePopupShell(VERSION, 
			 applicationShellWidgetClass, parent, args, cnt);

  /* create toplevel pane widget */
  cnt = 0;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNspacing, 0); cnt++;

  l->listener_window = XtCreateManagedWidget("pane", 
		   xmPanedWindowWidgetClass, l->listener_shell, args, cnt);

  /* create the menubar and load menus */
  cnt = 0;
  XtSetArg(args[cnt], XmNallowResize, True); cnt++;
  XtSetArg(args[cnt], XmNskipAdjust, False); cnt++;
  XtSetArg(args[cnt], XmNresizeHeight, True); cnt++;
  l->menubar = 
    exCreateMenuBar(l->listener_window, args, cnt, insertMenuReturnValue);
		
  if (makeFullName(getMenuDir(), ESH_MENUS, fullName, sizeof(fullName)) < 0) 
    return((Listener *) 0);

  if (! initMenus(l->menubar, l->listener_window, fullName, 
		  			l->popupMenus, &(l->popupCnt))) {
    sprintf(msg, "Cannot load menus from file -- %s --", fullName);
    exAlert(parent, -1, -1, msg);
  }

  /* create view editor */
  if (cols > 0 && rows > 0) {
    view->lines = rows / 3 * 2;
    command->lines = rows - view->lines;
    view->cols = cols;
    command->cols = view->cols - PROMPTWIDTH;
  }
  
  cnt = 0;
  l->view = makeEditor(l->listener_window, "view", view, args, cnt);

  bf_view = bufinfoptr(l->view);

  /* font related values can be provided now */
  cnt = 0;
  XtSetArg(args[cnt], XmNpaneMaximum, 50 * bf_view->fh); cnt++;
  XtSetArg(args[cnt], XmNpaneMinimum, 2 * bf_view->fh); cnt++;
  XtSetValues(bf_view->w, args, cnt);


  /* create form to contain prompt window and command editor */
  cnt = 0;
  XtSetArg(args[cnt], XmNmarginHeight, 0); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, 0); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, 0); cnt++;
  l->form_window = XtCreateManagedWidget("form", xmFormWidgetClass,
					   l->listener_window, args, cnt);

  /* create the prompt window */
  cnt = 0;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, 0); cnt++;
  XtSetArg(args[cnt], XmNuserData, l); cnt++;

  l->prompt = XtCreateManagedWidget("prompt", xmBulletinBoardWidgetClass,
			   l->form_window, args, cnt);

  XtOverrideTranslations(l->prompt, ptt);

  /* create command editor */
  cnt = 0;
  XtSetArg(args[cnt], XmNrightAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNtopAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNbottomAttachment, XmATTACH_FORM); cnt++;
  XtSetArg(args[cnt], XmNleftAttachment, XmATTACH_WIDGET); cnt++;
  XtSetArg(args[cnt], XmNleftWidget, l->prompt); cnt++;
  l->command = makeEditor(l->form_window, "command", command, args, cnt);

  /* set the font related values of the form window */
  bf_command = bufinfoptr(l->command);
  
  cnt = 0;
  XtSetArg(args[cnt], XmNpaneMaximum, 20 * bf_command->fh); cnt++;
  XtSetArg(args[cnt], XmNpaneMinimum, 5 * bf_command->fh); cnt++;
  XtSetValues(l->form_window, args, cnt);


  /* get the font related to the prompt window */

  cnt = 0;
  XtSetArg(args[cnt], XmNtextFontList, &fl); cnt++;
  XtGetValues(l->prompt, args, cnt);

  if (! fl) {
    l->font = exBoldFont(l->display);
    exError("Create listener", "cannot obtain prompt font!", NULL);
  } else 
    l->font= getFont(fl);

  /* set prompt window width */
  cnt = 0;
  XtSetArg(args[cnt], XmNwidth, PROMPTWIDTH * exFontWidth(l->font)); cnt++;
  XtSetValues(l->prompt, args, cnt);

  /* create the message area */
  cnt = 0;
  h = 3 / 2 * MAX(bf_view->fh, bf_command->fh);

  XtSetArg(args[cnt], XmNpaneMaximum, h); cnt++;
  XtSetArg(args[cnt], XmNpaneMinimum, h); cnt++;
  XtSetArg(args[cnt], XmNheight, h); cnt++;
  w = XtCreateManagedWidget("msg", xmDrawingAreaWidgetClass,
					 l->listener_window, args, cnt);
  
  editorSetMsg(l->view, w);
  editorSetMsg(l->command, w);
  setViewMode(l->view); 
  return(l);
}

/* openShell: ****************************************************************
 * fork a shell-process and connect it via pty 
 ****************************************************************************/

openShell(l, lines, cols, slot, shell, sh)
     Listener *l;
     int cols, lines;				/* window size */
     int slot;					/* in application array */
     char *shell, *sh;
{ 
  struct ptydesc pt;
  int pid, i;
  extern char *getenv(), *rindex();

  signal(SIGTTOU, SIG_IGN);
  
  if (openpty(&pt) < 0) 
    fatal("openShell", "cannot open connection (pty); terminated!", NULL);

  ttyGet();
  
  setWinSize(pt.pt_pfd, lines, cols);		/* tell the window size */
  
  if ((pid = vfork()) < 0)
    fatal("openShell", "cannot fork shell; terminated!", NULL);
  
  if (pid == 0) {				/* the son */
    int pgrp = getpid();
    
    if (removeCtrlTerminal() < 0) 
      fatal("openShell", "cannot start --", shell, 
	    " -- (Problems with control terminal)", NULL);
      
    close(pt.pt_pfd);				/* close unused end */

    ttySet(pt.pt_tfd); 
    
    dup2(pt.pt_tfd, 0);				/* establish communicat. */
    dup2(pt.pt_tfd, 1);
    dup2(pt.pt_tfd, 2);
    
    for (i = 3; i < getdtablesize(); i++)
      close(i);
    
    adjustEnvironment(lines, cols, shell);
    
    signal(SIGTERM, SIG_DFL);

    /* form a new precess group and associate the pty as control terminal */
    ioctl(0, TIOCSPGRP, &pgrp);			/* associate process group */
    setpgrp(0, 0);
    close(open(ttyname(0), O_WRONLY, 0));	/* make pty ctrl. terminal */
    setpgrp(0, pgrp);				/* create new process group */

    signal(SIGCHLD, SIG_DFL);
    signal(SIGHUP,  SIG_IGN);
    signal(SIGINT,  SIG_DFL);
    signal(SIGQUIT, SIG_DFL);
    signal(SIGTSTP, SIG_DFL);
    
    execlp(shell, sh, (char *) 0);
    fatal("openShell", "cannot exec shell; terminated!", NULL);
  }
  l->application[slot].procName = strCopy(shell);
  l->application[slot].procLine = strCopy(shell);
  l->application[slot].fd = pt.pt_pfd;
  l->application[slot].pid = pid;
  l->application[slot].flags = ESH_SYS | ESH_CONNECT | 
    					ESH_OCCUPIED | ESH_STARTED;
  
  adjustReadMask(l);

  /*sprintf(msg, "inferior shell -- %s -- started (pid=%d, fd=%d) !", 
	  					shell, pid, pt.pt_pfd);
  exError("ESH", msg, NULL);*/
  return(0);
}

/* clearPrompt: ***************************************************************
 * clear the esh prompt window 
 *****************************************************************************/

clearPrompt(l)
     Listener *l;
{
  XClearWindow(l->display, XtWindow(l->prompt));
}

/* writePrompt: ************************************************************
 * write the prompt into the esh-specific pronmpt window 
 ***************************************************************************/
/*ARGSUSED*/
static void writePrompt(w, event, params, num_params)
     Widget w;
     XLeaveWindowEvent *event;
     String *params;
     Cardinal *num_params;
{  
  Listener *l = NULL;
  Arg args[2];
  int cnt;

  cnt = 0;
  XtSetArg(args[cnt], XmNuserData, &l); cnt++;
  XtGetValues(w, args, cnt);

  if (! l)
    return;

  doWritePrompt(l);
  displayDate(l);
}

doWritePrompt(l)
     Listener *l;
{
  static char prompt[20];
  sprintf(prompt, "esh %d: ", l->cmdIndex);
  XSetFont(l->display, l->gc, l->font->fid);

  XDrawImageString(l->display, XtWindow(l->prompt), l->gc, 
	      5, exFontHeight(l->font) / 2 + l->font->max_bounds.ascent, 
		   prompt, strlen(prompt));
}

/* eshMsg: ********************************************************************
 * forward a message to the user. If a msg-window has been defined, the 
 * message is displayed there, otherwise a notifier is displayed.
 *****************************************************************************/

eshMsg(m)
     char *m;
{
  if (listener)
    editorMsg(listener->command, m);
  else
    exError("ESH", msg, NULL);
}

/* openApplication: ***********************************************************
 * fork the application as a child of the esh; returns the pid of the child 
 * process
 ****************************************************************************/
	
openApplication(l, slot, procName, parms)
     Listener *l;
     int slot;
     char *procName;		/* application name */
     char *parms[];		/* parameters */
{
  struct ptydesc pt;
  int pid, i;
  struct sgttyb ttyp;

  l->application[slot].pid = 0;
  pt.pt_pfd = -1;
  pt.pt_tfd = -1;

  if (l->application[slot].flags & ESH_CONNECT) {
    if (openpty(&pt) < 0)
      return(-1);
    
    /* set cbreak-mode for both ends of the pty */
    if (ioctl(pt.pt_pfd, TIOCGETP, &ttyp) < 0)
      return(-1);
    ttyp.sg_flags |= RAW;
    ttyp.sg_flags &= ~ECHO;
    if (ioctl(pt.pt_pfd, TIOCSETP, &ttyp) < 0)
      return(-1);
    if (ioctl(pt.pt_tfd, TIOCGETP, &ttyp) < 0)
      return(-1);
    ttyp.sg_flags |= RAW;
    if (ioctl(pt.pt_tfd, TIOCSETP, &ttyp) < 0)
      return(-1);
  }

  if ((pid = vfork()) < 0)    
    return(-1);

  if (pid == 0) {		/* the son */
    signal(SIGINT, SIG_IGN);
    
    if (removeCtrlTerminal() < 0) 
      fatal("openApplication", "cannot start -- ", 
	    procName, " -- (problems with the control terminal)", NULL);

    if (l->application[slot].flags & ESH_CONNECT) {
      dup2(pt.pt_tfd, 0);		
      dup2(pt.pt_tfd, 1);
    }

    for (i = 3; i < getdtablesize(); i++)
      close(i);
    
    execvp(procName, parms);

    sprintf(msg, "cannot start (%d) application -- %s --\n", errno,procName);
    fatal("ESH", msg, NULL);
  }

  l->application[slot].pid = pid;
  l->application[slot].flags |= ESH_STARTED;

  if (l->application[slot].flags & ESH_CONNECT) {
    l->application[slot].fd = pt.pt_pfd;
    close(pt.pt_tfd);
    adjustReadMask(l);
  } else
    l->application[slot].fd = 0;

  /* sprintf(msg, "application -- %s -- started (pid=%d, fd=%d)!", 
	                                       procName, pid, pt.pt_pfd);
  exError("ESH", msg, NULL);*/
  return(slot);
}

/* initApplications: **********************************************************
 * initialize the application structure with the system applications, namely 
 * history tracer, interface editor and the application shell. Then process 
 * the init file and add the user requested components to the application 
 * structure.
 *****************************************************************************/

Boolean initApplications(l, run_all) 
     Listener *l;
     Boolean run_all;
{
  char *parms[ESH_MAXPARMS];
  char line[EX_MAXSTR];

  char *shellName, *shName, *getenv();

  if (!(shellName = getenv("SHELL"))) {
    shellName = SHELLNAME;
    shName = SHNAME;
  }
  else {
    shName = rindex(shellName, '/');
    shName++;
  }

  openShell(l, editorLines(l->view), editorCols(l->view), 
	    				SHELL, shellName, shName);

  if (run_all) {	  	   
    parms[0] = IFEDITOR;	           /* start interface editor */
    parms[2] = NULL;
    l->application[IFEDIT].flags = ESH_CONNECT;
    if (openApplication(l, IFEDIT, IFEDITOR, parms) != -1){
      l->application[IFEDIT].procName = strCopy(IFEDITOR);
      sprintf(line, "+!%s", IFEDITOR);
      l->application[IFEDIT].procLine = strCopy(IFEDITOR);
      l->application[IFEDIT].flags |= ESH_OCCUPIED;
    }
  }             /* do not start ifeditor with option -n */

  parms[0] = HTRACER;			/* start history tracer */
  parms[2] = NULL;
  l->application[HTRACE].flags = ESH_CONNECT;
  if (openApplication(l, HTRACE, HTRACER, parms) != -1) {
    l->application[HTRACE].procName = strCopy(HTRACER);
    sprintf(line, "+!%s", HTRACER);
    l->application[HTRACE].procLine = strCopy(HTRACER);
    l->application[HTRACE].flags |= ESH_OCCUPIED;
  }

  return(processInitFile(l));
}

/* processInitFile: ***********************************************************
 * process the init file '.eshrc' if existing.
 *****************************************************************************/

Boolean processInitFile(l)
     Listener *l;
{
  FILE *f;
  char *home;
  char *getenv();
  char buf[256];
  int slot;

  if (! (home = getenv("HOME")))
    return(False);

  if (makeFullName(home, ".eshrc", buf, sizeof(buf)) < 0)
    return(False);

  if (! (f = fopen(buf, "r")))
    return(False);

  while (fgets(buf, sizeof(buf), f)) {
    if (buf[0] == '\n')			/* empty line */
      continue;
    buf[strlen(buf) - 1] = EOS;
    if ((slot = getFreeApplicationSlot(l)) != -1) 
      processApplicationLine(l, slot, buf);
  }
  
  fclose(f);
  return(True);
}
    
/* processApplicationLine: ****************************************************
 * process one line of the init file  and start the corresponding 
 * applications.
 * Each line in the init file has the following syntax:
 * [flag] application [parameters]
 * where flag is none or one of the following characters:
 *     '+'     to indicate that the application should be started 
 * 	       automatically by the command listener
 *     '!'     to indicate that a pipe should be installed between the 
 *             command listener and this application
 *     '$'     indicates a system application, that cannot be handled by the 
 *             user 
 *****************************************************************************/

processApplicationLine(l, slot, line)
     Listener *l;
     int slot;
     char *line;
{
  Boolean dostart = False;
  char *parms[ESH_MAXPARMS];
  register char *ptr;
  char *procName;
  register i, cnt, rc;
  char *strtok(), *rindex();
  
  if (*line == '+') {			/* start application */
    line++;
    dostart = True;
  }
  
  if (*line == '!') {			/* establish connection */
    l->application[slot].flags |= ESH_CONNECT;
    line++;
  }

  if (*line == '$') {			/* system application */
    l->application[slot].flags |= ESH_SYS;
    line++;
  }

  /* parse the string containing the application name and eventually the 
     parameters for the call 
  */
  
  for (i = 0, ptr = strtok(line, " "); ptr; ptr = strtok(NULL, " "))
    parms[i++] = ptr;
  
  cnt = i;
  parms[cnt] = NULL;
  
  procName = strCopy(parms[0]);	      	/* full application name */

  if (strcmp(procName, HTRACER) == 0) 
    slot = HTRACE;
  else if (strcmp(procName, IFEDITOR) == 0)
    slot = IFEDIT;
	  
  if (ptr = rindex(parms[0], '/'))
    parms[0] = ++ptr;

  if (! l->application[slot].procName)
    l->application[slot].procName = strCopy(parms[0]);
  if (! l->application[slot].procLine)
    l->application[slot].procLine = strCopy(line);

  l->application[slot].flags |= ESH_OCCUPIED;

  if (dostart) 
    rc = openApplication(l, slot, procName, parms);

  free(procName);
  return(rc);
}

/* getFreeApplicationSlot: ****************************************************
 * get a free slot in the listeners application table.
 *****************************************************************************/

getFreeApplicationSlot(l)
     Listener *l;
{
  register i;
  
  for (i = 0; i < ESH_MAXAPPL; i++) {
    if (! (l->application[i].flags & ESH_OCCUPIED)) {
      return(i);
    }
  }
  
  return(-1);
}

/* sendMsg: ******************************************************************
 * send a message to the interface editor, if it has been started. The 
 * message is specified by the pid, opc and info parameters.
 *****************************************************************************/

sendMsg(l, pid, opc, info)
     Listener  *l;
     int pid, opc, info;
{
  if (! (l->application[IFEDIT].flags & ESH_STARTED))
    return(-1);
  
  /*sprintf(msg, "sending msg: pid=%d, opc=%d, info=%d", pid, opc, info);
  exError("ESH", msg, NULL);*/
  return(exSendMsg(l->application[IFEDIT].fd, pid, opc, info));
}

/* sendName: *****************************************************************
 * send a character string (normally the name of a started application to the 
 * interface editor
 ***************************************************************************/

sendName(l, pid, name)
     Listener  *l;
     int pid;
     char *name;
{
  if (! (l->application[IFEDIT].flags & ESH_STARTED))
    return(-1);

  if (! name)
    return(-1);

  /*sprintf(msg, "sending name: pid=%d, name=-%s-", pid, name);
  exError("ESH", msg, NULL);*/
  return(exSendName(l->application[IFEDIT].fd, pid, name));
}

/* initSystemMenus ********************************************************
 * initialise the MDS structures and store the system application into the
 * the MDSarray. 
 *************************************************************************/

Boolean initSystemMenus(menubar, parent)
     exMenubar *menubar;
     Widget parent;
{
  int err;
  char fullName[256];

    if (makeFullName(getMenuDir(), SYSMENUS, fullName, 256) < 0) {
      return(False);
    }
    if ((err = restoreFromFile(parent, fullName, -1)) != ERR_NOERR) {
      uim_error(err);
      return(False);
    }
  return(loadMenus(menubar, parent, currentApplication(), 
		   (MenuDescr *)NULL, (int *)0));
}

/* initialise: ***************************************************************
 * do some initialisation
 *****************************************************************************/

static void initialise(d)
     Display *d;
{
  static Boolean initialised = False;

  if (initialised)
    return;

  normal_gc = XCreateGC(d, XDefaultRootWindow(d), 0, 0);
  XSetBackground(d, normal_gc, exWhitePixel(d));
  XSetForeground(d, normal_gc, exBlackPixel(d));
  XSetFunction(d, normal_gc, GXcopy);
  XSetFillStyle(d, normal_gc, FillSolid);
  XSetLineAttributes(d, normal_gc, 1, LineSolid, CapButt, JoinMiter);
  
  initialised = True;
}

/* insertMenuReturnValue: *****************************************************
 * callback that is called when the user has finished a menu selection. The 
 * return value is fetched and inserted into the command buffer 
 *****************************************************************************/

/*ARGSUSED*/
static void insertMenuReturnValue(w, event, params, num_params)
     Widget w;
     XPropertyEvent *event;
     String *params;
     Cardinal *num_params;
{
  register char *str;

  str = getMenuEventReturnValue(event);
  if (! str || ! *str)
    return;
  
  editorInsert(listener->command, str);
  free(str);
}

/* closeFunctionKeySet: ******************************************************
 * remove the current function key set from the screen 
 *****************************************************************************/

/*ARGSUSED*/
closeFunctionKeySet(parms)
     RequestParm *parms;
{
  closeFKSet();
}

/* loadFunctionKeySet: *******************************************************
 * load a set of function keys from a file and display them on the screen
 *****************************************************************************/

loadFunctionKeySet(parms)
     RequestParm *parms;
{
  loadFKSet(parms->parent, insertFunctionKeyValue);
}

/* show_eshell: **************************************************************
 * pop the eshell up
 ****************************************************************************/

show_eshell(l)
     Listener *l;
{
  XtPopup(l->listener_shell, XtGrabNone);
}

/* hide_eshell: **************************************************************
 * pop the eshell down
 ****************************************************************************/

hide_shell(l)
     Listener *l;
{
  XtPopdown(l->listener_shell);
  
  /* inform interface editor */

  if (l->application[IFEDIT].fd > 0) 	/* IFE running */
    sendMsg(l, getpid(), IFE_CLOSE, 0); 
}

/* insertFunctionKeyValue: **************************************************
 * this function is called whenever the user presses the left mouse button 
 * within a function key area. This function is provided when a 
 * function-key set is loaded.
 ****************************************************************************/

/*ARGSUSED*/
static void insertFunctionKeyValue(w, event, params, num_params)
     Widget w;
     XPropertyEvent *event;
     String *params;
     Cardinal *num_params;
{  
  register char *ptr;

  if (! (ptr = getFKValue(event)))
    return;

  editorInsert(listener->command, ptr);

  free(ptr);
}

/* eshSetSizeHints: **********************************************************
 * set some resize hints for the window manager 
 ****************************************************************************/

eshSetSizeHints(l)
  Listener *l;
{
  BufInfo *bf_command, *bf_view;
  XSizeHints size_hints;
  
  bf_view = bufinfoptr(l->view);
  bf_command = bufinfoptr(l->command);
  
  size_hints.flags = PMinSize | PResizeInc;
  size_hints.min_width  = 
    PROMPTWIDTH * exFontWidth(l->font) + 10 * bf_command->fw;
  size_hints.min_height =  
    menuBarHeight(l->menubar) + 2 * bf_view->fh + 2 * bf_command->fh;

  size_hints.width_inc  = bf_command->fw;
  size_hints.height_inc = bf_command->fh + 2 * FONTGAP;

  XSetNormalHints(l->display, XtWindow(l->listener_shell), &size_hints);
}
