/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include "fm.h"
#include <signal.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <grp.h>
#include <pwd.h>
#include <fcntl.h>
#include <errno.h>
#include <strings.h>
  
#include <Xm/Xm.h>

extern char *strrchr();
void setTitle();
static char msg[255];
time_t getStatusChangeTime();

/* calcPath: ****************************************************************
 * calculate an absolute file-name from an absolute path and a file name.
 * the files . and .. are handled correctly; also the case, when the file-name 
 * already is in absolute notation is handled. The resulting pathname is copied
 * into the variable path;
 ****************************************************************************/

calcPath(pathname,file)
     char *pathname, *file;
{
  char *ptr;

  if (strcmp(file,".") == 0) 
    return;
  
  if (*file == '/') { /* absolue file name given */
    strcpy(pathname,file);
    return;
  }
  
  if (strcmp(file,"..") == 0) { 
    if (strcmp(pathname,"/") == 0) 	/* already root */
      return;
    
    ptr = strrchr(pathname,'/');
    if (ptr == (char *) 0) 		/* impossible pathname */
      fatal("illegal path","calcPath","path: ",pathname,NULL);
    *ptr = '\0';
    
    if (strlen(pathname) == 0) 
      strcpy(pathname,"/");
    return;
  }
  
  if (pathname[strlen(pathname) - 1] != '/')
    strcat(pathname,"/");
  strcat(pathname,file);
}

/* renameDirWindows: ********************************************************
 * find all displayed windows with the name of the given directory window as 
 * a prefix of their name. For all such directories, change the path-component
 * to the vale given in 'newname'.
 ****************************************************************************/

renameDirWindows(curw, item, newname)
     dirWindow *curw;
     dirItem *item;
     char *newname;
{
  register dirWindow *w;
  static char dirname[PATHLEN];
  register char *name;
  register len;
  register char *rptr;
  
  strcpy(dirname, curw->path);
  strcat(dirname, "/");
  
  len = strlen(dirname);
  
  for (w = Head->first; w; w = w->next) {	/* loop through dir-windows */
    if (strncmp(dirname, w->path, len) == 0) { 	/* basename found */
      /* now consider two cases:
	 1. the item-name to be replaced is the last component
	 2. the item-name to be replaced is in-between
	 */
      if (! strcmp(w->path + len, ITEM_NAME(item))) {	/* last component */
	name = Malloc(len + strlen(newname) + 1);
	strcpy(name, dirname);
	strcat(name, newname);
      } else {						/* in-between */
	if ((rptr = index(w->path + len, '/')) == (char *) NULL)
	  continue;
	*rptr = '\0';
	if (strcmp(w->path + len, ITEM_NAME(item))) {	/* revert */
	  *rptr = '/';
	  continue;
	} else {					/* replace */
	  *rptr = '/';
	  name = Malloc(len + strlen(newname) + strlen(rptr) + 1);
	  strcpy(name, dirname);
	  strcat(name, newname);
	  strcat(name, rptr);
	}
      }
      free(w->path);
      w->path = name;
    }
  }
}      

/* lookifDisplayed: **********************************************************
 * check all displayed directory windows whether one exists with the given
 * name; in order to check, the name is compared with the original UNIX file-
 * name of the directory; if a window is found, a pointer to the 
 * directory-window structure is returned, NULL otherwise.
 ****************************************************************************/

dirWindow *lookifDisplayed(path)
     char *path;
{
  register dirWindow *w;
  
  for (w = Head->first; w; w = w->next) {
    if (strcmp(w->u_path, path) == 0) {
      return(w);
    }
  }
  return((dirWindow *) NULL);
}

/* lookupWindowByName: *******************************************************
 * check all displayed directory windows whether one exists with the given
 * name; in order to check, the name is compared with the probably renamed 
 * name of the directory; if a window is found, a pointer to the 
 * directory-window structure is returned, NULL otherwise.
 ****************************************************************************/

dirWindow *lookupWindowByName(name)
     char *name;
{
  register dirWindow *w;
  
  for (w = Head->first; w; w = w->next)
    if (strcmp(name, w->path) == 0)
      return(w);
  
  return((dirWindow *) NULL);
}

/* checkParentWindow: ********************************************************
 * look if the father-directory of the directory with the name given in path 
 * is already displayed; if so set the appropriate flag and redisplay the 
 * corresponding item;
 *****************************************************************************/

checkParentWindow(path, on)
     char *path;		/* path name of directory */
     Boolean on;		/* set (reset) the item_displayed flag */
{
  register dirWindow *w;
  register dirItem *iptr;
  char *cptr;
  char buf[PATHLEN];

  strcpy(buf, path);
  if (! (cptr = rindex(buf, '/')))		/* name of parent directory */
    return;
  
  *cptr++ = EOS;     				/* item in parent dir. */
  
  for (w = Head->first; w ; w = w->next) {
    if (strcmp(buf, w->u_path) == 0) {	/* parent directory has been found */
      for (iptr = w->itemList; iptr; iptr = iptr->next) {
	if (strcmp(ITEM_NAME(iptr),cptr) == 0) {/* item of current dir */
	  if (on)
	    iptr->flags |= ITEM_DISPLAYED;     	/* mark item as displayed */
	  else 
	    iptr->flags &= ~ITEM_DISPLAYED;    	/* mark item as not displayed */
	  fmdisplayItem(iptr->dirWnd, iptr, True);
	  return;
	}
      }
    }
  }
}

/* updateTree: ***************************************************************
 * look if the directory is already been displayed in a tree. If so mark the 
 * corresponding node (hilite it).
 *****************************************************************************/

void updateTree(path, hilite)
     char *path;		/* path to be displayed */
     Boolean hilite;		/* if True, hilite node */
{
  register dirWindow *w;
  register dirNode *nptr;
  int p_len, len;
  int cnt;
  Arg args[2];

  p_len = strlen(path);

  for (w = Head->first; w ; w = w->next) {	/* check each displ. dir. */
    if (! w->tree)				/* no tree created yet */
      continue;

    if (p_len <= (len = strlen(w->u_path)))	/* cannot be a valid prefix */
      continue;

    if (strncmp(w->u_path, path, len) == 0) {	/* valid prefix */
      for (nptr = w->tree->nodes; nptr; nptr = nptr->next) {
	if (strcmp(nptr->name, path) == 0) {	/* found the node */
	  cnt = 0;
	  if (hilite) {
	    nptr->flags |= NODE_SELECTED;   
	    XtSetArg(args[cnt], XmNforeground, fm_white_pixel); cnt++; 
	    XtSetArg(args[cnt], XmNbackground, fm_black_pixel); cnt++; 
	  } else { 
	    nptr->flags &= ~NODE_SELECTED;
	    XtSetArg(args[cnt], XmNforeground, fm_black_pixel); cnt++; 
	    XtSetArg(args[cnt], XmNbackground, fm_white_pixel); cnt++; 
	  } 
	  XtSetValues(nptr->node, args, cnt);
	  nptr->flags |= NODE_SELECTED;
	  break;				/* there's at most one node */
	}
      }
    }
  }
}

/* sys_copy: *****************************************************************
 * generate a command to copy or move the specified files from the 
 * directory 'source' to the destination directory; the command is sent to the 
 * shell for execution. The following command is generated:
 * cd <source>; cp (or mv) <f1> <dest>/<f2>
 ****************************************************************************/

sys_copy(source, f1, dest, f2, dflag)
     char *source, *f1, *dest, *f2;
     Boolean dflag;
{
  char command[255];
  
  sprintf(command, "cd %s; %s %s %s/%s", 
	  source, dflag ? "mv" : "cp", f1, dest, f2);
  system(command);
}

/* displayAttributes: ********************************************************
 * display the attributes of the files in the specified directory in a special
 * purpose editor 
 ****************************************************************************/

displayAttributes(wnd, editor)
     dirWindow *wnd;
     Editor editor;
{
  static char *perm_chars = "ugsrwxrwxrwx";
  int i;
  char line[100], buf[100], gname[15], oname[15], tbuf[15];
  register dirItem *iptr;
  u_short mode;
  struct group *g;
  struct passwd *pw;
  char *memset();
  
  clearBuffer(editor);
  
  sprintf(line, 
     "Files in directory -- %s -- on device -- ??? --\n\n\n", wnd->u_path);
  editorAppend(editor, line);
  
  sprintf(line, 
     "inode permissions   lnks  ( group  ,owner   )     size    file-name\n");
  editorAppend(editor, line);
  sprintf(line, 
     "date of last access  <|> date of last modification <|> date of last change\n"); 
  editorAppend(editor, line);
  sprintf(line, "___________________________________________________________________________\n\n");
  editorAppend(editor, line);
  
  for (iptr = wnd->itemList; iptr; iptr = iptr->next) { 
    SET_ITEM_STATE(wnd, iptr);
    if (iptr->flags & ITEM_SELECTED) {
      getStat(iptr, wnd->u_path, iptr->D_name);
      mode = iptr->sbp->st_mode;
      switch (mode & S_IFMT) {
      case S_IFIFO:
	buf[0] = 'p';
	break;
      case S_IFCHR:
	buf[0] = 'c';
	break;
      case S_IFDIR:
	buf[0] = 'd';
	break;
      case S_IFBLK:
	buf[0] = 'b';
	break;
      case S_IFREG:
	buf[0] = '-';
	break;
      case S_IFLNK:
	buf[0] = 'l';
	break;
      case S_IFSOCK:
	buf[0] = 's';
	break;
      default:
	buf[0] = '?';
	break;
      }
      
      for (i = 0; i < 12; i++)
	buf[i + 1] = (mode & (1 <<  (11 - i))) ? perm_chars[i] : '-';
      
      buf[13] = EOS;
      
      if (pw = getpwuid(iptr->sbp->st_uid))
	sprintf(oname, "%10.10s", pw->pw_name);
      else
	sprintf(oname, "?%8.8d?", iptr->sbp->st_uid);
      
      if (g = getgrgid(iptr->sbp->st_gid))
	sprintf(gname, "%10.10s", g->gr_name);
      else
	sprintf(gname, "?%8.8d?", iptr->sbp->st_gid);
      
      i = 0;
      memset(line, ' ', 100);				/* clear line */
      sprintf(tbuf, "%d", iptr->sbp->st_ino);		/* inode */
      bcopy(tbuf, line + 5 - strlen(tbuf), strlen(tbuf));
      bcopy(buf, line + 6, strlen(buf));		/* permissions */
      i = 6 + strlen(buf) + 1;
      sprintf(tbuf, "%d", iptr->sbp->st_nlink);		/* # of links */
      bcopy(tbuf, line + i + 2 - strlen(tbuf), strlen(tbuf));
      i += 3;
      bcopy("(", line + i, 1);
      i++;
      bcopy(oname, line + i + 10 - strlen(oname), strlen(oname));
      i += 10;
      bcopy(",", line + i, 1);
      i++;
      bcopy(gname, line + i + 10 - strlen(gname), strlen(gname));
      i += 10;
      bcopy(")", line + i, 1);
      i += 2;
      sprintf(tbuf, "%d", iptr->sbp->st_size);
      bcopy(tbuf, line + i + 7 - strlen(tbuf), strlen(tbuf));
      i += 9;
      bcopy(ITEM_NAME(iptr), line + i, strlen(ITEM_NAME(iptr)));
      line[i + strlen(ITEM_NAME(iptr))] = '\n';
      line[i + 1 + strlen(ITEM_NAME(iptr))] = EOS;
      editorAppend(editor, line);
      
      memset(line, ' ', 100);				/* clear line */
      
      sprintf(buf, "%s", asctime(localtime(&iptr->sbp->st_atime)));
      bcopy(&buf[11], &line[0], 8);			/* copy time */
      bcopy(&buf[8], &line[9], 2);			/* day of date */
      line[11] = '.';
      buf[7] = EOS;
      sprintf(tbuf, "%d", convertMonth(&buf[4]));	/* month of date */
      bcopy(tbuf, &line[14 - strlen(tbuf)], strlen(tbuf));
      line[14] = '.';
      bcopy(&buf[20], &line[15], 4);			/* year of date */
      bcopy("<|>", &line[20], 3);
      
      i = 24;
      sprintf(buf, "%s", asctime(localtime(&iptr->sbp->st_mtime)));
      bcopy(&buf[11], &line[i], 8);			/* copy time */
      bcopy(&buf[8], &line[i + 9], 2);			/* day of date */
      line[i + 11] = '.';
      buf[7] = EOS;
      sprintf(tbuf, "%d", convertMonth(&buf[4]));	/* month of date */
      bcopy(tbuf, &line[i + 14 - strlen(tbuf)], strlen(tbuf));
      line[i + 14] = '.';
      bcopy(&buf[20], &line[i + 15], 4);		/* year of date */
      bcopy("<|>", &line[i + 20], 3);
      
      i = 48;
      sprintf(buf, "%s", asctime(localtime(&iptr->sbp->st_mtime)));
      bcopy(&buf[11], &line[i], 8);			/* copy time */
      bcopy(&buf[8], &line[i + 9], 2);			/* day of date */
      line[i + 11] = '.';
      buf[7] = EOS;
      sprintf(tbuf, "%d", convertMonth(&buf[4]));	/* month of date */
      bcopy(tbuf, &line[i + 14 - strlen(tbuf)], strlen(tbuf));
      line[i + 14] = '.';
      bcopy(&buf[20], &line[i + 15], 4);		/* year of date */
      
      line[i + 19] = '\n';
      line[i + 20] = '\n';
      line[i + 21] = EOS;
      
      editorAppend(editor, line);
      
    }
  }
}

int convertMonth(str)
     char *str;
{
  register i;
  static char *months[] = {"Jan", "Feb", "Mar", "Apr", "May", "Jun", 
			     "Jul", "Aug", "Sep", "Oct", "Nov", "Dec"};
  
  for (i = 0; i < 12; i++)
    if (strcmp(str, months[i]) == 0)
      return(i + 1);
  
  return(-1);
}

/* tail: *********************************************************************
 * This function returns a pointer into path of the last node in the path 
 *****************************************************************************/

char *tail(path)
     char *path;
{
  char *ptr;

  if ((ptr = strrchr(path,'/')) == NULL)
    return(path);      		/* it didn't have a path seperator */
  else
    return(ptr + 1);   		/* return pointer to last node in path */
}

/* tableInsert: ***************************************************************
 * insert the string at the given index into the table; the entry is freed
 *****************************************************************************/

tableInsert(table, count, string)
     char *table[];
     int count;
     char *string;
{
  free(table[count]);
  table[count] = strCopy(string);
}

/* freeTable: *****************************************************************
 * frees the entries of the given table that holds 'max' entries
 *****************************************************************************/

freeTable(table, max)
char *table[];
int max;
{
  register i;

  for (i = 0; i < max; i++)
    free(table[i]);
}

/* initTable: *****************************************************************
 * initialises the given table with max entries
 *****************************************************************************/

initTable(table, max)
     char *table[];
     int max;
{
  register i;

  for (i = 0; i < max; i++)
    table[i] = (char *) NULL;
}

/* setFilePattern: ************************************************************
 * copy the file pattern into the dirWindow data-structure
 *****************************************************************************/

setFilePattern(w, p)
     dirWindow *w;
     char *p;
{
  char s[200];
  free(w->filePattern);
  w->filePattern = strCopy(p);
  
  sprintf(s, "%s/%s", w->path, p);
  setTitle(w->title, s);
}

/* setTitle: ****************************************************************
 * display the given string as new title; free resources of previous 
 * displayed string 
 ****************************************************************************/

void setTitle(title, str)
     Widget title;
     char *str;
{
  Arg args[1];
  XmString xmstr;
  int cnt;

  cnt = 0;
  XtSetArg(args[cnt], XmNlabelString, &xmstr); cnt++;
  XtGetValues(title, args, cnt);

  if (xmstr)
    XmStringFree(xmstr);

  cnt = 0;
  XtSetArg(args[cnt], XmNlabelString, 
	XmStringCreate(str, XmSTRING_DEFAULT_CHARSET)); cnt++; 
  XtSetValues(title, args, cnt);
}

/* getStat: ******************************************************************
 * get the stat-record for the specified file and store it with the dir item 
 ****************************************************************************/
getStat(item, pathname, file)
     dirItem *item;
     char *pathname, *file;
{
  char *name = (char *) malloc(strlen(pathname) + strlen(file) + 2);
  
  strcpy(name, pathname);
  strcat(name, "/");
  strcat(name, file);

  if (! item->sbp)
    item->sbp = (struct stat *) malloc(sizeof(struct stat));

  if ((stat(name, item->sbp)) == -1) {
    sprintf(msg, "Cannot stat file: %s",name);
    exAlert(Head->root, 100, 100, msg);
    item->flags |= ITEM_ERROR;
  }
  free(name);
}

/* isDirItem: ****************************************************************
 * checks if the specified item denotes a directory 
 ****************************************************************************/

Boolean isdirItem(iptr)
     dirItem *iptr;
{
      return(((iptr->sbp->st_mode & S_IFMT) == S_IFDIR) ? True : False);
}


/* setReadTime: **************************************************************
 * set the time when the directory has been read; this time is used to check
 * whether the information stored in the directory window is still valid.
 *****************************************************************************/

setReadTime(w)
     dirWindow *w;
{
  w->readTime = MAX(time(0), getStatusChangeTime(w->u_path));
}

/* getStatusChangeTime: ******************************************************
 * get the time from the stat-record of a file that defines when the file has 
 * been changed for the last time.
 ****************************************************************************/

time_t getStatusChangeTime(file)
     char *file;
{
  struct stat sbp;

  if (stat(file, &sbp) == -1) {
    sprintf(msg, "Cannot stat file: %s",file);
    exAlert(Head->root, 100, 100, msg);
    return(0L);
  }
  return(sbp.st_ctime);
}

/* makeItemRepresentation: ****************************************************
 * generates a printable representation of the item and its associated flags 
 *****************************************************************************/

char *makeItemRepresentation(iptr)
     dirItem *iptr;
{
  static char name[MAXITEMLEN + 1];
  char *nptr;
  
  if (! iptr)
    return((char *) NULL);
  
  name[0] = (iptr->flags & ITEM_ERROR) ? '?': ' ';
  
  name[1] = (iptr->flags & ITEM_DISPLAYED)? '!' : ' ';
  
  name[2] = (iptr->flags & ITEM_DELETED) ? 'D' : ' ';
  
  name[3] = (iptr->flags & ITEM_RENAMED) ? 'R' : ' ';
  
  name[4] = (iptr->flags & ITEM_ISDIR) ? '*' : ' ';
  
  name[5] = ' ';
  
  nptr = ITEM_NAME(iptr);
  
  strncpy(name + ITEMFLAGLEN, nptr, ITEMLEN);
  strncpy(name + strlen(nptr) + ITEMFLAGLEN, "                              ", 
	  			MAXITEMLEN - strlen(name));
  
  return(name);
}

/* titleHeight: **************************************************************
 * calculate the actual height of title widget
 ****************************************************************************/

titleHeight(tw)
     Widget tw;
{
  Arg args[10];
  int cnt;
  Dimension mbot, mtop, mh, bw, st;
  XmFontList fl;
  XmString xstr;

  if (! tw)
    return(0);

  cnt = 0;
  XtSetArg(args[cnt], XmNfontList, &fl); cnt++;
  XtSetArg(args[cnt], XmNlabelString, &xstr); cnt++;
  XtSetArg(args[cnt], XmNmarginBottom, &mbot); cnt++;
  XtSetArg(args[cnt], XmNmarginTop, &mtop); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, &mh); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, &st); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;

  XtGetValues(tw, args, cnt);

  return(mbot + mtop + 2 * (st + bw + mh) + XmStringHeight(fl, xstr));
}

/* mbWidth: **************************************************************
 * calculate the actual width of the menubar
 ****************************************************************************/

mbWidth(mb)
     exMenubar *mb;
{
  Arg args[10];
  int i, cnt, width;
  Dimension mw, bw, st, space, dbst, sad, hlt;
  XmFontList fl;
  XmString xstr;

  if (! mb)
    return(0);

  /* retrieve values from the menubar's rowcolumn widget */
  
  cnt = 0;
  mw = bw = st = 0;
  XtSetArg(args[cnt], XmNshadowThickness, &st); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, &mw); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;
  XtSetArg(args[cnt], XmNspacing, &space); cnt++;
  XtGetValues(mb->menubar, args, cnt);

  width = 2 * (mw + bw + st);

  /* retrieve values from button if any */
  cnt = 0;
  mw = bw = st = 0;

  XtSetArg(args[cnt], XmNdefaultButtonShadowThickness, &dbst); cnt++;
  XtSetArg(args[cnt], XmNshowAsDefault, &sad); cnt++;
  XtSetArg(args[cnt], XmNhighlightThickness, &hlt); cnt++;

  XtSetArg(args[cnt], XmNshadowThickness, &st); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;
  XtSetArg(args[cnt], XmNmarginWidth, &mw); cnt++;
  XtSetArg(args[cnt], XmNfontList, &fl); cnt++;
  XtSetArg(args[cnt], XmNlabelString, &xstr); cnt++;

  i = 0;
  while (mb->buttons[i].button) {
    XtGetValues(mb->buttons[i].button, args, cnt);
    width += 2 * (st + bw + mw + dbst + sad + hlt) 
                             + XmStringWidth(fl, xstr) + space;
    i++;
  }

  return(width);
}

/* sepHeight: **************************************************************
 * calculate the actual height of separator widget
 ****************************************************************************/

sepHeight(s)
     Widget s;
{
  Arg args[10];
  int cnt;
  Dimension marg, st, bw;

  if (! s)
    return(0);

  cnt = 0;
  XtSetArg(args[cnt], XmNmargin, &marg); cnt++;
  XtSetArg(args[cnt], XmNshadowThickness, &st); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;

  XtGetValues(s, args, cnt);

  return(marg + 2 * (st + bw));
}

/* updateDirectories: ********************************************************
 * this function is called from time to time to check if any directory window
 * needs to be updated; the interval for checking is determined by the 
 * constant UPDATE_INTERVAL
 ****************************************************************************/

updateDirectories()
{
  register dirWindow *w;

  for (w = Head->first; w; w = w->next) {
    if (needUpdate(w))
      updateDirWindow(w);
    alarm(UPDATE_INTERVAL);
  }
}
