/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include "exMsg.h"
#include <stdio.h>

#ifdef DEBUG
#include <mnemosyne.h>
#endif

static struct exMSG m;

/**************************************************************************
 *                      Public Functions
 ***************************************************************************/

/* exSendMsg ****************************************************************
 * Send a message between two connected components of the ERGO-shell to 
 * inform  the other component of the status or a performed action of the
 * first component.
 ***************************************************************************/ 

exSendMsg(fd, id, opc, info)
     int fd, id, opc, info;
{
  m.id = id;
  m.opcode = opc;
  m.info = info;
#ifdef DEBUG
  fprintf(stderr, "exSendMsg: fd=%d, id=%d, opc=%d, info=%d\n", 
	  fd, id, opc, info);
#endif
  if (write(fd, &m, sizeof(m)) != sizeof(m))
    return(-1);

  return(0);
}

/* exSendName ***************************************************************
 * Send a character string (e.g. the name of an application) to another
 * process.
 ***************************************************************************/ 

exSendName(fd, id, name)
     int fd, id;
     char *name;
{
  if (! name)
    return(-1);

  m.id = id;
  m.opcode = EXMSG_NAME;
  m.info = strlen(name);

#ifdef DEBUG
  fprintf(stderr, "exSendName: fd=%d, id=%d, opc=%d, info=%d, name=%s\n", 
	  fd, id,  EXMSG_NAME, strlen(name), name);
#endif
  if (write(fd, &m, sizeof(m)) != sizeof(m))
    return(-1);
   
  if (write(fd, name, strlen(name)) != strlen(name))
    return(-1);

  return(0);
}

/* exReceiveMsg *************************************************************
 * Receive a message from a connected component of the ERGO-shell to get
 * information about the status or a performed action of the sending component.
 ***************************************************************************/ 

exReceiveMsg(fd, id, opc, info)
     int fd, *id, *opc, *info;
{
  if (read(fd, &m, sizeof(m)) != sizeof(m)) {
    *id = *opc = *info = -1;
    return(-1);
  }

  *id = m.id;
  *opc = m.opcode;
  *info = m.info;
#ifdef DEBUG
  fprintf(stderr, "exReceiveMsg: fd=%d, id=%d, opc=%d, info=%d\n", 
	  fd, *id, *opc, *info);
#endif
  return(0);
}


