/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/****************************************************************************
 * This file contains the functions to create notifier and entries
 ****************************************************************************/

#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/IntrinsicP.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/ShellP.h>
#include <Xm/Xm.h>
#include <Xm/MessageB.h>
#include <Xm/Text.h>
#include <exBasics.h>
#include <exLayout.h>
#include <cursors/attention>

static char SccsID[] = "@(#)exDialog.c	1.2\t11/4/92";            /* sccs keyword */

static int EXMSGANSWER = 0;

void exWidgetPos();

static void exYesButton();
static void exNoButton();
static void exOkButton();
static void exCancelButton();
static void executeDefaultButton();
static void initDialogTranslations();
static Cursor attentionCursor = None;

struct msgBoxStruct {
  Widget shell;
  Widget msgbox;
};

static struct msgBoxStruct *noteb = NULL;        /* holds notifier structure */
static struct msgBoxStruct *alertb = NULL;       /* holds alert structure */

static struct msgBoxStruct *initNotify();
static struct msgBoxStruct *initAlert();

/***************************************************************************
 * public routines
 **************************************************************************/

/* exNotifyUser: ************************************************************
 * print message in a notifier; return ANSWERYES if user chooses 'YES', 
 * ANSWERNO otherwise.
 ****************************************************************************/

exNotifyUser(parent, x, y, msg)
     Widget parent;
     int x, y;
     char *msg;
{
  XEvent event;
  Arg args[5];
  register cnt = 0;
  Widget button;

  if (!noteb)
    noteb = initNotify(parent);

  button = XmMessageBoxGetChild (noteb->msgbox, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild(button);
  
  XtSetArg(args[cnt], XmNmessageString, XmStringCreateLtoR(msg, 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetValues(noteb->msgbox, args, cnt);

  exWidgetPos(noteb->shell, x, y, 0, 0);

  XtPopup(noteb->shell, XtGrabExclusive);

  while (((TransientShellWidget)noteb->shell)->shell.popped_up) {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }

  return(EXMSGANSWER);
}

/* exNotify2User: ************************************************************
 * print message in a notifier; return ANSWERYES if user chooses 'YES', 
 * ANSWERNO if user chooses 'NO' and ANSWERCANCEL if 'CANCEL' is selected
 ****************************************************************************/

exNotify2User(parent, x, y, msg)
     Widget parent;
     int x, y;
     char *msg;
{
  XEvent event;
  Arg args[5];
  register cnt = 0;
  Widget button;

  if (!noteb)
    noteb = initNotify(parent);

  button = XmMessageBoxGetChild (noteb->msgbox, XmDIALOG_HELP_BUTTON);
  XtManageChild(button);

  XtSetArg(args[cnt], XmNmessageString, XmStringCreateLtoR(msg, 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetValues(noteb->msgbox, args, cnt);

  exWidgetPos(noteb->shell, x, y, 0, 0);

  XtPopup(noteb->shell, XtGrabExclusive);

  while (((TransientShellWidget)noteb->shell)->shell.popped_up) {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }

  return(EXMSGANSWER);
}

/* exAlert: *******************************************************************
 * print message in a notifier which can only be confirmed
 ****************************************************************************/
 
void exAlert(parent, x, y, msg)
     Widget parent;
     int x, y;
     char *msg;
{
  XEvent event;
  Arg args[5];
  register cnt = 0;

  if (!alertb)
    alertb = initAlert(parent);

  XtSetArg(args[cnt], XmNmessageString, XmStringCreateLtoR(msg, 
                                        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetValues(alertb->msgbox, args, cnt);

  exWidgetPos(alertb->shell, x, y, 0, 0);

  XtPopup(alertb->shell, XtGrabExclusive);

  while (((TransientShellWidget)alertb->shell)->shell.popped_up) {
    XtNextEvent(&event);
    XtDispatchEvent(&event);
  }
}

/* exAlert2: *****************************************************************
 * print two messages in a message box which can only be confirmed
 ****************************************************************************/

void exAlert2(parent, x, y, msg1, msg2)
     Widget parent;
     int x, y;
     char *msg1, *msg2;
{
  char *msg;

  msg = Malloc(strlen(msg1) + strlen(msg2) + 3);

  strcpy(msg,msg1);
  strcat(msg,msg2);
  exAlert(parent, x, y, msg);
}

/* exWidgetPos: *************************************************************
 * determine the widget position; if x and/or y < 0 position is determined 
 * cursor relative
 ***************************************************************************/

void exWidgetPos(shellW, x, y, def_width, def_height)
     Widget shellW;
     int x, y;
     int def_width, def_height;		/* default width and height */
{
  int mx, my, dw, dh;
  Dimension width, height, bw;
  Arg args[5];

  if (x < 0 || y < 0) {
    getMouse_xy(XtDisplay(shellW), &mx, &my);
  } else {
    mx = x;
    my = y;
  }

  XtSetArg(args[0], XmNwidth, &width);
  XtSetArg(args[1], XmNheight, &height);
  XtSetArg(args[2], XmNborderWidth, &bw);
  XtGetValues(shellW, args, 3);

  /* problem: if not explicitly specified, width and height of a widget 
   * are calculated from the child widgets when a widget is realized; 
   * if the function exWidgetPos is called before realizing a widget, 
   * width and height may still be 0, in which case it cannot be determined
   * if the widget exceeds the screen. Calling exWidgetPos after having 
   * realized the widget has the effect that the widget's outline is 
   * visible for moments on the default position (0,0) before it appears 
   * on the target position; 
   * decision: default height and width are used if not yet calculated; 
   */

  if (width <= 0 || height <= 0) { 
    width = def_width;
    height = def_height;
  }

  dw = XWidthOfScreen(XtScreen (shellW));
  dh = XHeightOfScreen(XtScreen (shellW));

  if (mx + width > dw)
    mx = dw - width - 2 * bw;
  if (my + height > dh)
    my = dh - height - 2 * bw;

  XtMoveWidget(shellW, mx, my);
}

/* Bell: ********************************************************************
 * ring the bell
 ****************************************************************************/

void Bell()
{
  putc(ctrl('g'), stderr);
}

/* Beep: *********************************************************************
 * ring the keyboard bell 
 ****************************************************************************/

void Beep(d)
     Display *d;
{
  XBell(d, 0);
}

/****************************************************************************
 * private routines 
 ****************************************************************************/

/* initNotify: *************************************************************
 * create widgets for message box
 *************************************************************************/

static struct msgBoxStruct *initNotify(parent)
     Widget parent;
{
  Arg args[20];
  XtCallbackRec call1[2], call2[2], call3[2];
  Widget exMsgBox, exMsgShell, button, label;
  register cnt = 0;
  Display *d = XtDisplay(parent);

  noteb = (struct msgBoxStruct *) Malloc(sizeof(struct msgBoxStruct));

  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
  XtSetArg(args[cnt], XmNallowShellResize, True); cnt++;
  exMsgShell = XtCreatePopupShell("notifyShell", transientShellWidgetClass, 
				  parent, args, (Cardinal) cnt);

  noteb->shell = exMsgShell;

  call1[0].callback = exYesButton;
  call1[0].closure = (caddr_t) noteb;
  call1[1].callback = NULL;
  call1[1].closure = NULL;

  call2[0].callback = exNoButton;
  call2[0].closure = (caddr_t) noteb;
  call2[1].callback = NULL;
  call2[1].closure = NULL;

  call3[0].callback = exCancelButton;
  call3[0].closure = (caddr_t) noteb;
  call3[1].callback = NULL;
  call3[1].closure = NULL;    

  cnt = 0;
  XtSetArg(args[cnt], XmNdialogType, XmDIALOG_MESSAGE); cnt++;
  XtSetArg(args[cnt], XmNokLabelString, XmStringCreate(" YES ", 
					XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNcancelLabelString, XmStringCreate(" NO ", 
				        XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNcancelCallback, call2); cnt++;
  XtSetArg(args[cnt], XmNhelpLabelString, XmStringCreate(" CANCEL ", 
					XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNhelpCallback, call3); cnt++;
  XtSetArg(args[cnt], XmNokCallback, call1); cnt++;
  XtSetArg(args[cnt], XmNtopShadowPixmap, exLight_greyPM(d));
  cnt++;
  XtSetArg(args[cnt], XmNbottomShadowPixmap, exGreyPM(d)); 
  cnt++;
  XtSetArg(args[cnt], XmNdefaultButtonType, XmDIALOG_OK_BUTTON); cnt++;
  XtSetArg(args[cnt], XmNmessageString, 
      XmStringCreateLtoR("No message specified\n", XmSTRING_DEFAULT_CHARSET));
  cnt++;

  exMsgBox = XtCreateManagedWidget("exNotify", xmMessageBoxWidgetClass, 
				   exMsgShell, args, (Cardinal) cnt);

  noteb->msgbox = exMsgBox;

  button = XmMessageBoxGetChild (exMsgBox, XmDIALOG_OK_BUTTON); 
  XtSetArg(args[0], XmNshadowThickness, 5); 
  XtSetValues(button, args, 1);

  button = XmMessageBoxGetChild (exMsgBox, XmDIALOG_CANCEL_BUTTON); 
  XtSetArg(args[0], XmNshadowThickness, 5);
  XtSetValues(button, args, 1);

  button = XmMessageBoxGetChild (exMsgBox, XmDIALOG_HELP_BUTTON); 
  XtSetArg(args[0], XmNshadowThickness, 5);
  XtSetValues(button, args, 1);

  label = XmMessageBoxGetChild (exMsgBox, XmDIALOG_SEPARATOR); 
  XtSetArg(args[0], XmNseparatorType, XmNO_LINE);
  XtSetValues(label, args, 1);

  initDialogTranslations(exMsgBox);

  return(noteb);
}

/* initAlert: *************************************************************
 * create widgets for message box
 *************************************************************************/

static struct msgBoxStruct *initAlert(parent)
     Widget parent;
{
  Arg args[20];
  XtCallbackRec call[2];
  Widget exMsgBox, exMsgShell, button, label;
  register cnt = 0;
  Display *d = XtDisplay(parent);

  alertb = (struct msgBoxStruct *) Malloc(sizeof(struct msgBoxStruct));

  XtSetArg(args[cnt], XmNkeyboardFocusPolicy, XmPOINTER); cnt++;
  XtSetArg(args[cnt], XmNallowShellResize, True); cnt++;
  exMsgShell = XtCreatePopupShell("notifyShell", transientShellWidgetClass, 
				  parent, args, (Cardinal) cnt);

  alertb->shell = exMsgShell;

  call[0].callback = exOkButton;
  call[0].closure = (caddr_t) alertb;
  call[1].callback = NULL;
  call[1].closure = NULL;

  cnt = 0;
  XtSetArg(args[cnt], XmNdialogType, XmDIALOG_ERROR); cnt++;
  XtSetArg(args[cnt], XmNokLabelString, XmStringCreate(" OK ", 
					XmSTRING_DEFAULT_CHARSET)); cnt++;
  XtSetArg(args[cnt], XmNokCallback, call); cnt++;
  XtSetArg(args[cnt], XmNtopShadowPixmap, exLight_greyPM(d)); cnt++;
  XtSetArg(args[cnt], XmNbottomShadowPixmap, exGreyPM(d)); 
  cnt++;
  XtSetArg(args[cnt], XmNdefaultButtonType, XmDIALOG_OK_BUTTON); cnt++;
  XtSetArg(args[cnt], XmNmessageString, 
      XmStringCreateLtoR("No message specified\n", XmSTRING_DEFAULT_CHARSET));
  cnt++;

  exMsgBox = XtCreateManagedWidget("exAlert", xmMessageBoxWidgetClass, 
				   exMsgShell, args, (Cardinal) cnt);

  alertb->msgbox = exMsgBox;

  button = XmMessageBoxGetChild (exMsgBox, XmDIALOG_OK_BUTTON); 
  XtSetArg(args[0], XmNshadowThickness, 5);
  XtSetValues(button, args, 1);

  button = XmMessageBoxGetChild (exMsgBox, XmDIALOG_CANCEL_BUTTON);
  XtUnmanageChild(button);
  button = XmMessageBoxGetChild (exMsgBox, XmDIALOG_HELP_BUTTON);
  XtUnmanageChild(button);

  label = XmMessageBoxGetChild (exMsgBox, XmDIALOG_SEPARATOR); 
  XtSetArg(args[0], XmNseparatorType, XmNO_LINE);
  XtSetValues(label, args, 1);

  initDialogTranslations(exMsgBox);

  return(alertb);
}

/* exYesButton: ***********************************************************
 * pops notifier down when YES button is pressed
 **************************************************************************/
/*ARGSUSED*/
static void exYesButton(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  struct msgBoxStruct *exNote_shell = (struct msgBoxStruct *) closure;

  EXMSGANSWER = ANSWERYES;

  exUngrabAll(exNote_shell->msgbox);

  XtPopdown(exNote_shell->shell);	
}

/* exNoButton: ************************************************************
 * pops notifier down when NO button is pressed
 **************************************************************************/
/*ARGSUSED*/
static void exNoButton(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  struct msgBoxStruct *exNote_shell = (struct msgBoxStruct *) closure;

  EXMSGANSWER = ANSWERNO;

  exUngrabAll(exNote_shell->msgbox);

  XtPopdown(exNote_shell->shell);	
}

/* exCancelButton: ************************************************************
 * pops notifier down when CANCEL button is pressed
 **************************************************************************/
/*ARGSUSED*/
static void exCancelButton(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  struct msgBoxStruct *exNote_shell = (struct msgBoxStruct *) closure;

  EXMSGANSWER = ANSWERCANCEL;

  exUngrabAll(exNote_shell->msgbox);

  XtPopdown(exNote_shell->shell);	
}

/* exOkButton: ***********************************************************
 * pops notifier down when OK button is pressed
 **************************************************************************/
/*ARGSUSED*/
static void exOkButton(widget, closure, callData)
     Widget widget;
     caddr_t closure;
     caddr_t callData;
{
  struct msgBoxStruct *exNote_shell = (struct msgBoxStruct *) closure;

  exUngrabAll(exNote_shell->msgbox);

  XtPopdown(exNote_shell->shell);	
}

/* grabDialog: ***************************************************************
 * grab pointer for dialog components
 ****************************************************************************/
/*ARGSUSED*/
static void grabDialog(w, event, params, num_params)
     Widget w;
     XKeyEvent *event;
     String *params;
     Cardinal *num_params;
{
  exGrabAll(w, attentionCursor);
}

/* initDialogTranslations: *************************************************
 * add translations in order to actively grab pointer when a dialog 
 * component is exposed
 **************************************************************************/

static void initDialogTranslations(w)
     Widget w;
{
  static Boolean INIT = False;
  static XtTranslations ptt;
 
  static XtActionsRec dialogCmds[] = {
    {"grabDialog", grabDialog},
    {"executeDefaultButton", executeDefaultButton}
  };
  static char dialogCmdTable[] = 
    "<Expose>		: grabDialog() \n \
     <Key>Return	: executeDefaultButton()";

  if (!INIT) {
    Pixmap pm;
    XColor fg, bg;
    
    XtAddActions(dialogCmds, 2);
    ptt = XtParseTranslationTable(dialogCmdTable);
    pm = XCreateBitmapFromData (XtDisplay(w), 
				XDefaultRootWindow(XtDisplay(w)), 
				attention_bits, 
				attention_width, attention_height);
    attentionCursor = XCreatePixmapCursor(XtDisplay(w), pm, pm, &fg, &bg, 1,1);
    INIT = True;
  }

  XtOverrideTranslations(w, ptt);
}

/* executeDefaultButton: *****************************************************
 * translation function that is called when the user presses the return-key 
 * when a notifier is displayed. As the result the default button action 
 * routine is executed.
 ****************************************************************************/
/*ARGSUSED*/
static void executeDefaultButton(w, event, params, num_params)
     Widget w;
     XKeyEvent *event;
     String *params;
     Cardinal *num_params;
{
  Widget button, btn;

  if (! (button = XmMessageBoxGetChild(w, XmDIALOG_DEFAULT_BUTTON)))
    return;

  if (! (btn = XmMessageBoxGetChild(w, XmDIALOG_OK_BUTTON)))
    return;

  if (button == btn)
    XtCallCallbacks(w, XmNokCallback, NULL);
  else
    XtCallCallbacks(w, XmNcancelCallback, NULL);
}

/* getMenuDir: ***************************************************************
 * retrieve the value of the environment variables ESHBASEDIR and HOSTTYPE to
 * build the name of the directory where the system menus are stored.
 * This path must have the following structure:
 * $(ESHBASEDIR/include/menus
 ****************************************************************************/

char *getMenuDir()
{
  char *ptr;
  char fullName[256];
  char *getenv();

  if (!(ptr = getenv("ESHBASEDIR"))) 
    fatal("getMenuDir", 
	  "Environment variable ESHBASEDIR not defined: exiting !", NULL);
	
  makeFullName(ptr, "include/menus", fullName, 256);

  return(strCopy(fullName));
}

/* exGrabAll: ***************************************************************
 * grab pointer and keyboard with respect to the given widget 
 ****************************************************************************/

void exGrabAll(w, csr)
     Widget w;
     Cursor csr;
{
  if (w) {
    XtGrabPointer(w, False, 
	   ButtonPressMask | ButtonReleaseMask | PointerMotionMask, 
	   GrabModeAsync, GrabModeAsync, None, csr, CurrentTime);

    XtGrabKeyboard(w, False, GrabModeAsync, GrabModeAsync, CurrentTime);
  }
}

/* exUngrabAll: ***************************************************************
 * grab pointer and keyboard with respect to the given widget 
 ****************************************************************************/

void exUngrabAll(w)
     Widget w;
{
  if (w) {
    XtUngrabPointer(w, CurrentTime);
    XtUngrabKeyboard(w, CurrentTime);
  }
}
