/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
#include <X11/IntrinsicP.h>
#include <Xm/Xm.h>

#include <exBasics.h>
#include "exEdit.h"

#define UP 1
#define DOWN 0

#define META             0x80

char *EMtext();

/* exCursorPos: **************************************************************
 * set the current position according to the mouse cursor; additionally the
 * dot is set to the current position
 ****************************************************************************/

/*ARGSUSED*/
void exCursorPos(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  BufInfo *bf;
  Arg args[1];

  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);
  
  if (bf->region.m_line >= 0) {
    ESdisplayRegion(bf, False);
    reset_region(bf);
  }
  cursorPos(bf, event->x / exFontWidth(bf->font), 
	        event->y / LINEHEIGHT(bf->font));
  setDot(bf, event->x / exFontWidth(bf->font), event->y / LINEHEIGHT(bf->font));
}

/* exCursorSelect: ***********************************************************
 * set the current position according to the mouse cursor; additionally the
 * mark is set if a dot has been defined. The region is highlighted.
 ****************************************************************************/

/*ARGSUSED*/
void exCursorSelect(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  BufInfo *bf;
  Arg args[1];

  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);

  cursorSelect(bf, event->x / exFontWidth(bf->font), 
	       		event->y / LINEHEIGHT(bf->font));
}

/* exCursorDeselect **********************************************************
 * deselect text area and reset region 
 *****************************************************************************/

/*ARGSUSED*/
void exCursorDeselect(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  BufInfo *bf;
  Arg args[1];

  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);

  cursorDeselect(bf);
}

/* exCursorCopy: **************************************************************
 * Copy the region into the internal kill buffer and the cut-buffer
 ****************************************************************************/

/*ARGSUSED*/
void exCursorCopy(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  BufInfo *bf;
  Arg args[1];
  
  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);
  
  cursorCopy(bf);
}

/* exCursorCut: **************************************************************
 * Copy the region into the internal kill buffer and the cut-buffer; the 
 * delete the region.
 ****************************************************************************/

/*ARGSUSED*/
void exCursorCut(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  BufInfo *bf;
  Arg args[1];
  
  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);

  cursorCut(bf);
}

/* exCursorPaste: ************************************************************
 * Paste the contents of the kill buffer to the current position
 ****************************************************************************/

/*ARGSUSED*/
void exCursorPaste(w, event, params, num_params)
     Widget w;
     XButtonEvent *event;
     String params;
     int *num_params;
{
  BufInfo *bf;
  Arg args[1];

  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);

  cursorPaste(bf, event->x / exFontWidth(bf->font), 
	        		event->y / LINEHEIGHT(bf->font));
}

/* gkbd: **********************************************************************
 * provide  the next input character to the edit-routine; 
 *****************************************************************************/

gkbd(bf, edStr, len, chr)
     BufInfo *bf;
     char *edStr;
     int len;
     char *chr;
{
  if(bf->edPtr == len) {
    bf->edPtr = 0;
    *chr = EOS;
    return;
  }
  if (edStr[bf->edPtr] & META) {		/* meta bit mappd to ESC */
    edStr[bf->edPtr] &= ~META;			/* reset META bit */
    *chr = ESC;
  } else
    *chr = edStr[bf->edPtr++];
}

/* activateEditor: ***********************************************************
 * activates or deactivates the editor by changing the cursor appearance; an 
 * activated editor has a solid cursor, a deactivated editor a cursor frame; 
 * is usually called, when the mouse enters or leave the editor window;
 *****************************************************************************/

void activateEditor(bf, on)
     BufInfo *bf;
     Boolean on;
{
  switch (on) {
  case True:
    bf->flags |= ED_ACTIVE;      
    PSshowframe(bf, False);                            	/* hide frame */
    ESshowcur(bf, True, bf->curcol, bf->currow);	/* draw cursor */
    break;
  case False:
    ESshowcur(bf, False, bf->curcol, bf->currow);	/* hide cursor */
    if (! (bf->flags & (ED_FRAMEOFF | ED_CUROFF))) {
      storeFrame(bf, bf->curcol, bf->currow);
      PSshowframe(bf, True);                            /* draw frame */	
    }                        
    bf->flags &= ~ED_ACTIVE;				
    break;
  }
}

/* exKey: *******************************************************************
 * get the next characters from the keyboard 
 ***************************************************************************/
/*ARGSUSED*/
void exKey(w, event, params, num_params)
     Widget w;
     XKeyEvent *event;
     String params;
     int *num_params;
{
  char cmd[1024];
  BufInfo *bf;
  Arg args[1];
  int cnt;

  XtSetArg(args[0], XmNuserData, &bf);
  XtGetValues(w, args, 1);

  if (event->state & Mod1Mask) {	/* meta key pressed */
    cmd[0] = ESC; 
    cmd[1] = EOS;
    edit(bf, cmd, 1);
  }

  cnt = (* bf->inputFnct)(event, cmd, sizeof(cmd), NULL, NULL);
  
  if (cnt > 0) {
    if (cnt < sizeof(cmd)) 
      cmd[cnt] = EOS;
    else 
      cmd[sizeof(cmd) -1 ] = EOS;
    edit(bf, cmd, cnt);
  }
}

/* cursorPos: *****************************************************************
 * position the cursor to the given position
 *****************************************************************************/

cursorPos(bf, x, y)
     BufInfo *bf;
     int x, y;
{
  int l, sl, i, w, xl, maxLine = ETnumlines(bf);

  /* find corresponding phys. line */

  y = MIN(bf->lines - 1, y);

  /* if screen partially filled, check if in empty area */

  for (l = 0, i = bf->topline; i < maxLine; i++) {
    l += (sl = screenLines(bf, i));
    if (l > y)
      break;
  }
  
  if (i == maxLine) {		/* beyond the last line of the buffer */
    bf->curline = i - 1;
    bf->curpos = ESlen(bf, i - 1);
    phys2screen(bf, bf->curline, bf->curpos, &bf->currow, &bf->curcol, True);
  } else {			/* within valid text area */
    if (sl > 1) {	      	/* multi-line */
      w = ESwidth(bf);		/* screen width in characters */
      xl = y - (l - sl);	/* line-offset in multi-line */
      if (xl < sl - 1) 		/* not in last line of multiline */
	bf->curcol = MIN(x, w - 1);
       else 			/* in last line */
	bf->curcol = 		/* check rest of multi-line */
	  MIN(x, ESrestLen(bf, i) + 1);
    } else 			/* single line */
      bf->curcol = MIN(x, ESlen(bf, i));

    bf->currow = y;
    screen2phys(bf, bf->currow, bf->curcol, &bf->curline, &bf->curpos, False);
  }

  setpos(bf);
  bf->preservedCol = bf->curcol;
  ESsetcur(bf);
}

/* exKeyUp: *****************************************************************
 * this function is called when the user releases a key; the keyboard is 
 * checked and if no keys are presseed, all key-press events are cleared 
 * from the queue of pending events; thus, type-ahead is suppressed.
 ****************************************************************************/
/*ARGSUSED*/
/*
void exKeyUp(w, event, params, num_params)
     Widget w;
     XKeyEvent *event;
     String params;
     int *num_params;
{
  int d;
  XEvent e;
  char key_vector[32];

  XQueryKeymap(event->display, key_vector);
  
  for (d = 1; d < 32; d++)		* check if any key is pressed *
    if (key_vector[d]) 			* if so, return *
      return;
    

  while (XCheckTypedEvent(event->display, KeyPress, &e))
    ;					* eliminate pending key down events *
}
*/
/* cursorSelect: *************************************************************
 * set mark to current position and highlight the area between dot and mark
 ****************************************************************************/

cursorSelect(bf, x, y)
  BufInfo *bf;
  int x, y;
{
  cursorPos(bf, x, y);

  if (! setMark(bf, x, y))
    return;

  ESdisplayRegion(bf, True);
}


/* cursorDeselect: ***********************************************************
 * display the area between dor and mark in normal color and reset dot and mark
 ****************************************************************************/

cursorDeselect(bf)
     BufInfo *bf;
{
  ESdisplayRegion(bf, False);
  reset_region(bf);
}

/* cursorCopy: **************************************************************
 * copy the text between dot and mark into the kill-buffer and
 * make it available using the X-buffers
 ****************************************************************************/
cursorCopy(bf)
     BufInfo *bf;
{
  if (exCopy(bf) && bf->killBuff)
    XStoreBytes(XtDisplay(bf->w), bf->killBuff, strlen(bf->killBuff) + 1);
}

/* cursorCut: ****************************************************************
 * copy and delete the text between dot and mark into the kill-buffer and
 * make it available using the X-buffers
 ****************************************************************************/

cursorCut(bf)
     BufInfo *bf;
{
  if (exCut(bf) && bf->killBuff)
    XStoreBytes(XtDisplay(bf->w), bf->killBuff, strlen(bf->killBuff) + 1);
}

/* cursorPaste: **************************************************************
 * paste the contents of the X-buffer to the current position; additionally 
 * set the exEdit kill-buffer.
 ****************************************************************************/

cursorPaste(bf, x, y)
     BufInfo *bf;
     int x, y;
{
  int nbytes;
  char *s;

  if (bf->flags & ED_VIEWMODE) {
    ESmsg(bf, "Read only buffer!", False);
    return;
  }

  s = XFetchBytes(XtDisplay(bf->w), &nbytes);
    
  if (s && nbytes > 0)
    setKillBuffer(bf, strCopy(s));
  
  XFree(s);

  cursorPos(bf, x, y);

  exYank(bf);
}

