/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/******************************************************************************
 * This file contains the program to generate the macro editors menus;
 * the generated application node "macro-functions" is stored in file 
 * "macro.menu"
 *****************************************************************************/

#include <exBasics.h>
#include <exFonts.h>
#include <exErrors.h>
#include <exMenu.h>

/* options menu */
extern quitMacroEditor();


/* macro set menu */
extern openMacroSet();
extern closeMacroSet();
extern saveAsMacroSet();
extern newMacroSet();
extern deleteMacroSet();

/* macros menu */
extern newMacro();
extern saveMacro();
extern saveMacroInSet();
extern saveMacroAs();
extern loadMacro();
extern deleteMacros();

Widget top;

main(argc, argv)
     int argc;
     char **argv;
{
  char str[200];
  MHmenu *mptr;
  MHitem *iptr;
  MHiReq *req;

#ifdef DEBUG
  malloc_debug(2);
#endif

  top = XtInitialize(NULL, "fkm menus", (XrmOptionDescRec *) NULL, 0, &argc, 
		     argv);

  initMDS();
  exInitLayout(XtDisplay(top));

  newApplicationNode("macro-functions");
  
  insertFunction(currentApplication(), quitMacroEditor, "quitMacroEditor");
  insertFunction(currentApplication(), openMacroSet, "openMacroSet");
  insertFunction(currentApplication(), closeMacroSet, "closeMacroSet");
  insertFunction(currentApplication(), saveAsMacroSet, "saveAsMacroSet");
  insertFunction(currentApplication(), newMacroSet, "newMacroSet");
  insertFunction(currentApplication(), deleteMacroSet, "deleteMacroSet");

  insertFunction(currentApplication(), newMacro, "newMacro");
  insertFunction(currentApplication(), saveMacro, "saveMacro");
  insertFunction(currentApplication(), saveMacroAs, "saveMacroAs");
  insertFunction(currentApplication(), saveMacroInSet, "saveMacroInSet");
  insertFunction(currentApplication(), loadMacro, "loadMacro");
  insertFunction(currentApplication(), deleteMacros, "deleteMacros");
  
/* ---------------------------- options ------------------------------- */
  mptr = newMenu("quit");
  setMenuType(mptr, MDS_M_PULLD);
  setSelType(mptr,MDS_CNT_EQ);
  setSelCnt(mptr,1);
  sprintf(str, "Option to quit the macro editor.");
  mptr->m_hlp->hlpmsg = strCopy(str);

  iptr = newItem(mptr, "quit", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Quit the macro editor.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(quitMacroEditor, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);


/* ---------------------------- macro sets ------------------------------- */
  mptr = newMenu("macro sets");
  setMenuType(mptr, MDS_M_PULLD);
  setSelType(mptr,MDS_CNT_EQ);
  setSelCnt(mptr,1);
  sprintf(str, "Options to manage macro sets.");
  mptr->m_hlp->hlpmsg = strCopy(str);

  iptr = newItem(mptr, "load set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Load a macro set and provide the first macro in the buffer.\nYou get a menu containing the available macro sets\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(openMacroSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "new set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Create a new macro set.\nYou are requested to enter the new name for the set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(newMacroSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "save set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Save the current macro set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(closeMacroSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "save set as", EOS); 
  setRetPattern(iptr, "$ $");
  sprintf(str, "Duplicate the current macro set.\nYou are requested to enter the new name for the set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(reqViaWnd, "macro set name:", (char *) 0, 20);
  insertRequest (iptr, req, 0);                /* set request */
  req = makeOneRequest(saveAsMacroSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "delete set", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Delete the current macro set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(deleteMacroSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

/* ------------------------------- macro ------------------------------- */
  mptr = newMenu("macros");
  setMenuType(mptr, MDS_M_PULLD);
  setSelType(mptr,MDS_CNT_EQ);
  setSelCnt(mptr,1);
  sprintf(str, "Options to manage macros.");
  mptr->m_hlp->hlpmsg = strCopy(str);

  iptr = newItem(mptr, "load macro", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Load a macro from the current set.\nYou get a menu containing all macros in the current set\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(loadMacro, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "new macro", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Create a new macro in the current set.\nYou are requested to enter the name for the new macro.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(newMacro, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);
  
  iptr = newItem(mptr, "save macro", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Save the current macro into the current set.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(saveMacro, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);
 
  iptr = newItem(mptr, "save macro as", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Save the current macro under a different name\nin the current set.\nYou are prompted for the new name.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(saveMacroAs, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "save macro in" ,EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Save the current macro into a different set.\nYou get a menu containing all available sets\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(saveMacroInSet, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  iptr = newItem(mptr, "delete macros", EOS); 
  setRetPattern(iptr, "$");
  sprintf(str, "Delete macros from the current set.\nYou get a menu containing all macros in the current set\nfrom which you can select the desired one.");
  iptr->i_hlp->hlpmsg = strCopy(str);
  req = makeOneRequest(deleteMacros, "", "", 0);
  setReqCnt(req, 1, MDS_CNT_EQ);
  insertRequest(iptr, req, 0);

  setFile(currentApplication(), "macro.menu");
  dumpToFile(top);
  exit(0);
}



quitMacroEditor()
{
}


openMacroSet()
{
}

closeMacroSet()
{
}

saveAsMacroSet()
{
}

newMacroSet()
{
}

deleteMacroSet()
{
}



newMacro()
{
}

saveMacro()
{
}

saveMacroAs()
{
}

saveMacroInSet()
{
}

loadMacro()
{
}

deleteMacros()
{
}


