/************************* INCLUDES *************************************/

#include <strings.h>
#include <ctype.h>

/*#include <stdio.h>
#include <X11/Xlib.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <Xm/Xm.h>
#include <exBasics.h>
#include <exLayout.h>   in exMenuP.h included */

#include <exMenu.h>
#include <exFonts.h>
#include <exErrors.h>
#include <exEdit.h> 
#include <exEntry.h>
#include <exThermometer.h>

/*#include <events.h>*/
#include "editNames.h"

/************************* CONSTANT DEFINITIONS ************************/

#define PREVIEW_NAME    "Preview Window"

#define TFS		'\065'	 /* 'toggle font size' signal */

#define XPOS            80
#define YPOS            80
#define WIDTH           970
#define HEIGHT          515
#define BORDERWIDTH     2

#define MEDITOR_CLOSED  0x0      /* menu editor comes up as icon */
#define MEDITOR_OPEN    0x1      /* menu editor is completly shown */
#define ME_LARGEFONT    0x2      /* menu editor is displayed in large font */
#define XICON           1041     /* x-value for icon window */
#define YICON           110      /* y-value for icon window */

#define MAXROW 		100      /* maximum number of rows in text */
#define MAXCOL	 	200      /* maximum number of columns in text */
#define MAXHEIGTH 	2000     /* maximum heigth of non visible window */
#define MAXWIDTH 	3000     /* maximum width of non visible window */

#define LABELNO		10       /* number of static text strings in form */

#define ME_MAXITEM	32
#define MAXAPPL			MIN(MDS_MAXAPPL, ME_MAXITEM)
#define MAXMENU			MIN(MDS_MAXMENU, ME_MAXITEM)
#define MAXFUNC			MIN(MDS_MAXFUNC, ME_MAXITEM)

#define ALLOFF   	0x0      /* no other component active */
#define PREVIEWON       0x1      /* preview window open */
#define ACTEDON         0x2      /* activity mask editor on */
#define REQEDON		0x4	 /* request editor on */
#define HELPEDON	0x8	 /* help text editor on */

#define E_ENTRYLEN1	25 /*15*/      /* length of text entries */
#define E_ENTRYLEN2	3	 /* length of number entries */
#define E_ENTRYLEN3     40       /* length of return entries */
#define E_ENTRYBUFLEN   5        /* lines of return entries */

#define ME_INCOMPLETE	0x1      /* missing data in form */
#define ME_INFO         0x2      /* form info editor is mapped */

#define NOTELEN		1024     /* length of notice submitted to user */

#define WINMIN(a,b) (((a) < (b)) ? (a) : (b))
#define WINMAX(a,b) (((a) > (b)) ? (a) : (b))
#define SIZECALC(n,s,t) (((n) * (s)) / (t))


/********************* STRUCTURE DEFINITIONS ******************************/

/********** structure to describe windows needed for the menu editor *******/

typedef struct _meditWindow {
  Widget meShell;               	/* top level widget of menu editor */
  struct _exMenubar *menuBar;           /* menu bar of menu editor */
  Widget meForm;              		/* form widget of menu editor */
  struct _headerDescr *head;
  Widget meRowCol;                      /* row/column widget */
  struct _formDescr *form;   		/* structure describing the form of
					   menu editor */
  Widget scrollBar;   			/* scrollbar */   
  exMenu *preMenu;            		/* window of menu in preview mode */
  struct _previewWindow *pre_win;  	/* structure describing windows of 
					   previewer */
  int com_flag;              		/* previewer,activity mask editor, 
					   request editor on/off */
  int menu_flag;			/* if on special system menus have to
					   be recreated */
  int flag;                  		/* if on menu editor windows are 
					   mapped */
} meditWindow;

/**** structure to hold information needed for running the menu editor *****/

struct runMenuEdit {
  struct _meditWindow *menuEditDescr;       
  ApplicationNode *applPtr;         /* current application in menu editor */ 
  MHmenu *findPtr;                  /* menu which was last loaded in editor */
  MHmenu *menuPtr;                  /* copy of current menu to work on */
};

/************** structures to describe form header ***********************/

typedef struct _headerDescr {
  /*int width, height;                    * width and height in pixel */
  struct _formEntry *MenuEntry;
  Widget labelSet[LABELNO];             /* text in form window */
  Widget applForm;
  Widget setName[2];
  Widget setAccess[2];
  Widget systemMenu[2];
  XFontStruct *mFontPtr;       		/* basic font of menu editor */
  /*int flags;                            * consistent form, etc. */
} headerDescr;

/************** structures to describe form window ***********************/

typedef struct _formDescr {
  struct _formEntry *ItemEntries[ME_MAXITEM];
  struct _formEntry *ReturnEntries[ME_MAXITEM];
	/*  struct _formEntry *KeyEntries[ME_MAXITEM]; */
  Widget RequestLabels[ME_MAXITEM]; 
  int entryCnt;                          /* number of form entries */
  Editor formInfo;                       /* information about missing 
	                                    data etc. */
  int flags;                             /* consistent form, etc. */
} formDescr;

/************* structure to describing an entry in form ********************/

typedef struct _formEntry {
	Boolean valid;                	/* entry has value */
	char **string;              	/* holds editor contents */
	exEntry *entry;
} formEntry;

/************ structure to describe windows needed for the previewer *******/

typedef struct _previewWindow {
  Widget pShell;               /* popup-shell of previewer */
  Widget form;                 /* form of previewer */ 
  Widget text;                 /* label to display if a return values
				  is returned */
  Widget quitButton, repButton;
  Editor retShow;	       /* editor to display return value in */
  struct _exMenubar *menuBar;  /* menu bar of previewer */
} previewWindow;

/************************** EXTERNAL DEFINITIONS *************************/

extern struct runMenuEdit *rme;
extern MenuDescr EditMenuWin[];	/* system menu windows of menu editor */
extern int WinCnt;		/* number of system menu windows */

extern Boolean APPL_CHANGED;

extern char *ApplName[];	/* list of accessable application names */
extern char *MenuName[];     	/* list of menu names under current 
				   application */
extern char *FuncName[];	/* list of function names under 
				   current application */
extern char *ApplFuncName[];	/* list of application function names
				   only for current application */

extern XFontStruct *TitleFont, *BaseFont;
extern Display *display;
extern INTERNALS;
extern char msg[];              /* global array for messages */

/************************** EXTERNAL FUNCTIONS *************************/













extern ApplAccessType(/* parms */);
extern ApplChange(/* parms */);
extern ApplicationNode *createInitApplication(/* name */);
extern Boolean LoadEditMenus(/* w, mpath, mfile */);
extern Boolean ReplaceNotifier(/* aptr, mptr */);
extern Boolean acceptMenu(/* reminder */);
extern Boolean checkForm(/* meForm */);
extern Boolean checkFormEntry(/* fentry */);
extern Boolean checkMenuAccess(/* mptr */);
extern Boolean checkMenuReq(/* mptr */);
extern Boolean cmpFormEntry(/* fentry, string */);
extern Boolean createInternalMenus(/* parent */);
extern Boolean delMenuUserFunct(/* mptr */);
extern Boolean generateApplList(/* nameList */);
extern Boolean generateFuncList(/* aptr, funcList, applFunc */) ;
extern Boolean generateMenuList(/* aptr, nameList */);
extern MHmenu *findFirstMenu(/* aptr */);
extern MHmenu *generateListMenu(/* menuName, nameList, cnt, reqFunc, reqData, selTypeOne */);
extern MHmenu *initialMenu(/* mptr */);
extern MHmenu *saveMenu(/* meForm, mptr, editShow */);
extern Previewer(/* mEditWin, aptr, mptr */);
extern RestReq(/* mptr, returnEntry, iIndex */);
extern SaveChangesNotifier(/* aptr, mptr */);
extern SaveNewMenuNotifier(/* aptr */);
extern SystemApplFlag(/* parms */);
extern UserFunctNotifier(/* aptr, mptr */);
extern adjustDefaultAppl(/* aptr */);
extern adjustDefaultMenu(/* mptr */);
extern adminDuplMenu(/* parms */);
extern adminDuplSet(/* parms */);
extern adminNewMenu(/* parms */);
extern adminNewSet(/* parms */);
extern adminPreview(/* parms */);
extern adminQuit(/* parms */);
extern adminSave(/* parms */);
extern adminUserFunct(/* parms */);
extern adminWrite(/* parms */);
extern checkApplAccess(/* aptr */);
extern checkMenuPath(/* aptr */);
extern checkSubComp(/* editw */);
extern createHelpForm(/* mEditWin, mptr, x, y */);
extern defaultUserFunct(/* parms */);
extern delUserFunct(/* parms */);
extern dumpChanges(/* aptr */);
extern eNewline(/* editor */);
extern editAccessType(/* parms */);
extern editActivityMask(/* mptr, x, y */);
extern editMenuActMask(/* parms */);
extern editMenuHelp(/* parms */);
extern editMenuPos1(/* parms */);
extern editMenuPos2(/* parms */);
extern editMenuType(/* parms */);
extern editReqForm1(/* mptr, x, y */);
extern editRequest(/* parms */);
extern editSelCnt(/* parms */);
extern editSelType(/* parms */);
extern fillEditPopup(/* aptr, mptr */);
extern fillEditPulldown(/* mptr */);
extern fillEntry(/* fe, defStr, viewMode */);
extern fillFormWithAppl(/* editw, aptr */);
extern fillFormWithMenu(/* editw, mptr */);
extern formEntry *createFormEntry(/* parent, msg, len, buflin */);
extern genReqList(/* parms */);
extern getFormEntryHeight(/* fentry */);
extern listUserFunct(/* parms */);
extern main(/* argc, argv */);
extern meCopyMenu(/* parms */);
extern meDeleteApplication(/* parms */);
extern meDeleteMenu(/* parms */);
extern meDeleteRequest(/* parms */);
extern meSelectApplication(/* parms */);
extern meSelectMenu(/* parms */);
extern meSetInitialValue(/* mh, mds, item, req, initval */);
extern newApplList(/* parms */);
extern newFuncList(/* parms */);
extern newMenuList(/* parms */);
extern selectReqFunct(/* parms */);
extern stopPreview(/* parms */);
extern subCompState(/* mEditWin, index, on */);
extern void checkReqEdit(/* mEditWin */);
extern void quitReqEdit(/* parms */);
extern void resetSelection(/* mptr, menu */);
extern void updateReqCnt(/* editw, mptr, icnt */);
