/************************************************************************
 *                                                                      *
 *    Copyright (c) 1991 Wilhelm Valder, Regine Freitag, Wolfgang Dzida *
 *                        All rights reserved                           *
 *                                                                      *
 *                                                                      *
 *    Permission to copy this program or any part of this program is    *
 *    granted in case of personal, non-commercial use only. Any use     *
 *    for profit or other commercial gain without written permissions   *
 *    of the authors is prohibited.                                     *
 *                                                                      *
 *    You may freely distribute this program to friends, colleagues,    *
 *    etc.  As a courtesy we would appreciate it if you left our names  *
 *    on the code as the authors.                                       *
 *    We also, of course, disavow any liabilities                       *
 *    for any bizarre things you may choose to do with the system.      *
 *                                                                      *
 *    We would appreciate though your letting us know about any         *
 *    extensions, modifications, project suggestions, etc.              *
 *    Please mail all bugs, comments, complaints and suggestions to:    *
 *                                                                      *
 *                                                                      *
 *    GMD (German National Research Center for Computer Science)        *
 *    Institut fuer Systemtechnik (F2G2)                                *
 *    P.O. Box 1240                                                     *
 *    Schloss Birlinghoven                                              *
 *    D-5205 St. Augustin 1, West Germany                               *
 *                                                                      *
 *    email:    dzida@kmx.gmd.dbp.de                                    *
 *              freitag@gmdzi.uunet                                     *
 *                                                                      *
 ************************************************************************/
/****************************************************************************
 * This file contains utility functions 
 * - to handle the form components of the menu editor
 * - to create and check special applications and menus
 * - to give appropriate messages to the user
 ****************************************************************************/

#include "editDescr.h"
#include "formEntry.h"
#include <Xm/LabelG.h>

static int searchEntry();

/****************************************************************************
 * functions to handle the form components of the menu editor
 ****************************************************************************/

/* FillinEntry ***************************************************************
 * insert default string into entry and adapt view mode
 ***************************************************************************/

fillEntry(fe, defStr, viewMode)
     formEntry *fe;                         /* form entry to be filled in */
     char *defStr;                          /* default string */
     Boolean viewMode;
{
  if (defStr) {
    editorFillBuffer(fe->entry->entryEditor, defStr, False);
    fe->valid = TRUE;
  }
  else if (fe->valid) {
    clearBuffer(fe->entry->entryEditor);
    fe->valid = False;
  }

  switchViewMode(fe->entry->entryEditor, viewMode);
}

/* createFormEntry ***********************************************************
 * creates a form entry with label 'msg' and len 'len'
 * fuer scrollable entries muss zusaetzlich eine Laenge fuer den Text 
 * mitgegeben werden
 ****************************************************************************/

formEntry *createFormEntry(parent, msg, len, buflin)
     Widget parent;
     char *msg;                       /* entry label */
     int len;                         /* entry length in characters */
     int buflin;                      /* buffer lines for entry */            
{
  register formEntry *fe;

  fe = (formEntry *) Malloc (sizeof(formEntry));

  fe->valid = False;
  fe->string = (char **) Malloc(buflin * sizeof(char *));

  if (! (fe->entry = exMakeEntry(parent, fe->string, buflin,
				 len, msg, entryCmdTable, escTable, 
				 ctrlxTable, entryCmds, False, False)))
    {
    free(fe->string);
    free(fe);
    return((formEntry *) 0);
  }

  switchFrame(fe->entry->entryEditor, False);
  return(fe);
}

/* eNewline *****************************************************************
 * function to go to next entry on 'newline'-command (in editor)
 ****************************************************************************/

eNewline(editor)
     Editor editor;
{
 register eIndex = -1;
 Display *d;
 meditWindow *me;

 d = XtDisplay(rme->menuEditDescr->meShell);
 me = rme->menuEditDescr;
  
 if (editor == me->head->MenuEntry->entry->entryEditor)   
                                                   /* go to ItemEntries[0] */
   XWarpPointer(d, (Window) 0, editorWindow(me->form->ItemEntries[0]->entry->
					entryEditor), 0, 0, 0, 0, 0, 0);
 else if ((eIndex = searchEntry(editor, me->form->ItemEntries, 
				me->form->entryCnt)) != -1) 
                                             /* go to ReturnEntries[eIndex] */
   XWarpPointer(d, (Window) 0, editorWindow(me->form->ReturnEntries[eIndex]->
				      entry->entryEditor), 0, 0, 0, 0, 0, 0);
 else if ((eIndex = searchEntry(editor, me->form->ReturnEntries, 
				me->form->entryCnt)) != -1)
                                           /* go to ItemEntries[eIndex + 1] */
   XWarpPointer(d, (Window) 0, editorWindow(me->form->ItemEntries[eIndex + 1]->
				       entry->entryEditor), 0, 0, 0, 0, 0, 0);
}

/* searchEntry ***************************************************************
 * search for index in entry group given the editor window
 *****************************************************************************/

static int searchEntry(edit, entryGrp, grpSize)
     Editor edit;
     struct _formEntry *entryGrp[];
     int grpSize;
{
  register i;

  for (i = 0; i < grpSize; i++)
    if (entryGrp[i]->entry->entryEditor == edit)
      return(i);

  return(-1);
}

/* updateReqCnt **************************************************************
 * show number of requests which still have to be defined
 ****************************************************************************/

void updateReqCnt(editw, mptr, icnt)
     meditWindow *editw;
     MHmenu *mptr;
     int icnt;
{
  register i, cnt;
  char str[5];
  Arg args[5];

  for (i = 0; i < icnt; i++) {      /* request count */
    sprintf(str, "%d", RestReq(mptr, editw->form->ReturnEntries[i], i));
    cnt = 0;
    XtSetArg(args[cnt], XmNlabelString, XmStringCreateLtoR
	     (str, XmSTRING_DEFAULT_CHARSET)); cnt++;
    XtSetValues(editw->form->RequestLabels[i], args, cnt);
  }
}

/* RestReq *******************************************************************
 * calculate number of requests which still have to be defined
 ****************************************************************************/

RestReq(mptr, returnEntry, iIndex)
     MHmenu *mptr;
     formEntry *returnEntry;
     int iIndex;
{
  register rcnt = 0,              /* number of defined requests */
           cnt = 0;               /* number of requests still to be defined */

  if (iIndex > mptr->itemCnt)
    return(0);

  if (mptr && mptr->i_table[iIndex])     
    rcnt = mptr->i_table[iIndex]->i_return->reqCnt;

  if (returnEntry->valid) {
    cnt = countRequestsInPattern(returnEntry->string[0]);
    cnt -= rcnt;
  }

  if (cnt < 0)                    /* number of defined requests exceeds number
				     of '$' specified in return pattern */
    setError(ERR_VALUE);
  /* loesche request '?' ? */

  return(cnt);
}

/* acceptMenu ***************************************************************
 * checks if menu can be inserted in application node; returns 'TRUE' 
 * on success
 ****************************************************************************/

Boolean acceptMenu(reminder)
     Boolean reminder;                          /* if TRUE, asks user whether 
						   changes are to be saved */
{
  register answer = 0;

  if (! (saveMenu(rme->menuEditDescr, rme->menuPtr, TRUE))) {
    if (rme->menuEditDescr->form->flags & ME_INCOMPLETE)    
                                        /* incomplete menu is completed now */
      return(False);
    else                                     /* inconsistent menu not saved */
      return(TRUE);
  }

  if (rme->findPtr) {                            /* existing menu is edited */
    if (cmpMenuDescr(rme->menuPtr, rme->findPtr) == False)
      return(TRUE);                                  /* no changes occurred */
    else if (reminder) {
      answer = SaveChangesNotifier(rme->applPtr, rme->menuPtr);
      if (answer == ANSWERNO)
	return(TRUE);              /* changes in existing menu are not saved */
      else if (answer == ANSWERCANCEL)
	return(False);               /* cancel: return to current situation */
    }
  }

  if (!(rme->findPtr)) {                             /* new menu is defined */
    if (reminder) {
      answer = SaveNewMenuNotifier(rme->applPtr);
      if (answer == ANSWERNO) 
	return(TRUE);                              /* new menu is not saved */
      else if (answer == ANSWERCANCEL)
	return(False);               /* cancel: return to current situation */
    }

    if ((rme->findPtr = getMenuDescrByName(rme->menuPtr->m_repr->name)) &&
	(ReplaceNotifier(rme->applPtr, rme->findPtr) == False)) 
      return(TRUE);                   /* menu with this name already exists and
					 user does not want replacement */
    if (rme->applPtr->menuCnt >= MDS_MAXMENU) {
      exAlert(rme->menuEditDescr->meShell, -1, -1, 
	      "This menu set contains maximum number of menus; delete a menu to save current menu or save it in a different menu set");
      setError(ERR_MAX);
      return(False);
    }

    setCurrentApplication(rme->applPtr);   
    rme->findPtr = newMenu("@pointToMenu@");
    setWorkApplication();
  }

  copyMenuDescr(rme->menuPtr, rme->findPtr);
  generateMenuList(rme->applPtr, MenuName);
  adjustDefaultMenu(rme->menuPtr);
  rme->applPtr->_changed = TRUE;
  return(TRUE);
}  

/* saveMenu *****************************************************************
 * function to save edited menu (check consistency of form); returns 
 * empty menu pointer if it is an empty or inconsistent form
 ***************************************************************************/

MHmenu *saveMenu(meForm, mptr, editShow)
     struct _meditWindow *meForm;
     MHmenu *mptr;
     Boolean editShow;        /* if True, editor with information about
				 inconsistencies in menu definition is shown */
{
  XEvent event;
  register i, j = 0, k;
  char notice[NOTELEN];
  int x, y;
  Window subw;

  if (meForm->form->flags & ME_INFO)
    stxtClose(meForm->form->formInfo);
  if (!emptyBuffer(meForm->form->formInfo))
    clearBuffer(meForm->form->formInfo);

  if (! checkForm(meForm))                      /* empty form */
    return((MHmenu *) 0);

  if (meForm->head->MenuEntry->valid)
    mptr->m_repr->name = strCopy(meForm->head->MenuEntry->string[0]);
  else 
    appendText(meForm->form->formInfo, "No menu name defined\n");

  for (i = 0; i < meForm->form->entryCnt; i++) {

    if (meForm->form->ItemEntries[i]->valid 
	|| meForm->form->ReturnEntries[i]->valid) {
                                         /* item is (partially) defined */
      if (j >= mptr->itemCnt) {
	mptr->i_table[j] = newItem(mptr, NULL, EOS);
      } 

      if (meForm->form->ItemEntries[i]->valid)
	mptr->i_table[j]->i_repr->name = 
	  strCopy(meForm->form->ItemEntries[i]->string[0]);
      else 
	appendText(meForm->form->formInfo, 
		 sprintf(notice, "no item name defined for item %d\n", i + 1));

      if (meForm->form->ReturnEntries[i]->valid) {
	setRetPattern(mptr->i_table[j], 
		      meForm->form->ReturnEntries[i]->string[0]);
	if (RestReq(mptr, meForm->form->ReturnEntries[i], i) != 0)
	  appendText(meForm->form->formInfo, sprintf(notice,
                     "number of defined requests does not match number of $ in return pattern for item %d\n", i + 1));
      }
      else 
	appendText(meForm->form->formInfo, 
	     sprintf(notice,"no return pattern defined for item %d\n", i + 1));

      j++;			/* number of valid items */
    }
    else
      break;      
  }

  updateReqCnt(meForm, mptr, meForm->form->entryCnt);

  for (k = j; k < mptr->itemCnt; k++)
    deleteItem(mptr, mptr->i_table[k]);    /* if number of defined items is 
					      smaller than in former menu */

  if (editShow && !emptyBuffer(meForm->form->formInfo)) {   
                                                  /* missing data in form */
    
    getMouse_xy(XtDisplay(meForm->meShell), &x, &y);
    stxtShow(meForm->form->formInfo, x + 200, y + 200);
    meForm->flag |= (ME_INFO | ME_INCOMPLETE);

/*    while (XtPending()) {
      XtNextEvent(&event);
      XtDispatchEvent(&event);
    }
*/    
    if (exNotifyUser(meForm->meShell, -1, -1, "incomplete menu definitions cannot be saved;\ndo you want to go on editing this menu description\n(otherwise changes will be lost)?") == ANSWERNO)  {
      stxtClose(meForm->form->formInfo);
      meForm->flag &= ~(ME_INCOMPLETE | ME_INFO);     /* menu is lost */
      }

    return((MHmenu *) 0);  
  } 

  meForm->form->flags &= ~ME_INCOMPLETE;
  return(mptr);
}

/* checkForm ******************************************************************
 * marks every entry filled in and checks if form is empty
 *****************************************************************************/

Boolean checkForm(meForm)
     struct _meditWindow *meForm;
{
  register int i;
  Boolean valid = False;

  if (checkFormEntry(meForm->head->MenuEntry))
    valid = TRUE;
 
  for (i = 0; i < ME_MAXITEM; i++) {
    valid |= checkFormEntry(meForm->form->ItemEntries[i]);
    valid |= checkFormEntry(meForm->form->ReturnEntries[i]);
  }

  return(valid);
}

/* checkFormEntry ***********************************************************
 * determine if entry contents is valid; return TRUE if entry is valid
 ***************************************************************************/

Boolean checkFormEntry(fentry)
     formEntry *fentry;
{
  if (!fentry->entry) {
    setError(ERR_NULL);
    return(False);
  }

  if (fentry->string[0][0] != EOS)
    fentry->valid = TRUE;
  else
    fentry->valid = False; 

  return(fentry->valid);
}


/* cmpFormEntry ***********************************************************
 * if entry contents equals string return TRUE; False otherwise 
 ***************************************************************************/

Boolean cmpFormEntry(fentry, string)
     formEntry *fentry;
     char *string;
{
  if (!fentry->entry) {
    setError(ERR_NULL);
    return(False);
  }
 
  if (checkFormEntry(fentry)) {    	/* entry not empty */
    if (strCmp(string, getBufferContents2(fentry->entry->entryEditor)))
      return(False);
  }
  else            			/* entry empty */
    if (string)
      return(False);

  return(TRUE);
}


/* getFormEntryHeight ********************************************************
 * determine if entry contents is valid; return TRUE if entry is valid
 ***************************************************************************/

getFormEntryHeight(fentry)
     formEntry *fentry;
{
  register cnt = 0;
  Dimension eh, to, bw;
  Arg args[5];

  if (!fentry->entry) {
    setError(ERR_NULL);
    return(0);
  }
  
  XtSetArg(args[cnt], XmNheight, &eh); cnt++;
  XtSetArg(args[cnt], XmNmarginHeight, &to); cnt++;
  XtSetArg(args[cnt], XmNborderWidth, &bw); cnt++;
  XtGetValues(fentry->entry->entryWindow, args, cnt);

  return(eh + 2 * to + 2 * bw);
}

/****************************************************************************
 * functions to maintain applications and menus in menu editor
 ****************************************************************************/

/* dumpChanges ***************************************************************
 * dumps application node to file if application node or menus in it have 
 * changed
 ****************************************************************************/

dumpChanges(aptr)
     ApplicationNode *aptr;
{
  ApplicationNode *curPtr;
  int x, y;
  int initial = strcmp(aptr->applicationName, UNTITLEDAPPL);
  char *ptr;

  if (MenusToSave(aptr)) {
    curPtr = currentApplication();
    setCurrentApplication(aptr);

    aptr->_changed = False;

    if (initial == 0) {      /* application to store is initial application */
      getMouse_xy(XtDisplay(rme->menuEditDescr->meShell), &x, &y);
      while (emptyStr(aptr->applicationName) || 
	     strcmp(aptr->applicationName, UNTITLEDAPPL) == 0)
	aptr->applicationName = exGetEntry(rme->menuEditDescr->meShell, x, y, 
			      "give a name other than 'no name' to menu set:", 
					   (char *) 0, E_ENTRYNAME);
        
      if (strncmp(aptr->file, UNTITLEDAPPL, strlen(UNTITLEDAPPL)) == 0) {
	ptr = (char *) Malloc(strlen(aptr->applicationName) + 
			      strlen(MDS_FILESUFFIX));
	strcpy(ptr, aptr->applicationName);
	strcat(ptr, MDS_FILESUFFIX);
	setFile(aptr, ptr);
	fillFormWithAppl(rme->menuEditDescr, aptr);
	generateApplList(ApplName);
	free(ptr);
      }
    }
      
    checkMenuPath(aptr);
    dumpToFile(rme->menuEditDescr->meShell);

    setCurrentApplication(curPtr);  /* reset current application before dump */
  }
  else 
    if (initial == 0)
      removeApplNode(aptr);
}

/* createInitAplication ******************************************************
 * creates and inserts empty application
 ***************************************************************************/

ApplicationNode *createInitApplication(name)
     char *name;
{
  ApplicationNode *aptr;

  if (aptr = getApplByName(name)) {    
    sprintf(msg, "A menu set with name %s\n already exists. Select another name\n or delete existing menu set first.", name);
    exAlert(rme->menuEditDescr->meShell, -1, -1, msg);
    return((ApplicationNode *) 0);
  }      

  aptr = createApplNode();
  initApplNode(aptr, name);
  setAccess(aptr, MDS_CHNGACC);

  if (insertApplNode(aptr) == ERR_MAX) {
    sprintf(msg, "Maximum number of menu sets\nis exceeded; menu set %s\ncannot be created.", name);
    exAlert(rme->menuEditDescr->meShell, -1, -1, msg);
    return((ApplicationNode *) 0); 
  }

  return(aptr);
}

/* initialMenu ************************************************************
 * generates an initial menu to be edited in the menu editor
 ***************************************************************************/

MHmenu *initialMenu(mptr)
     MHmenu *mptr;
{
  if (mptr)
    clearMenu(mptr);
  
  mptr = createMenu();

  initMenu(mptr, MDS_M_PULLD, "");
  setMenuFlag(mptr, MDS_CHNGACC);

  return(mptr);
}

/* findFirstMenu ************************************************************
 * find first accessable menu in application node 
 ***************************************************************************/

MHmenu *findFirstMenu(aptr)
     ApplicationNode *aptr;
{
  register i;

  for (i = 0; i < aptr->menuCnt; i++) {    
    if (aptr->m_table[i]->m_flags & (MDS_READACC | MDS_CHNGACC))
      return(aptr->m_table[i]);
  }

  return((MHmenu *) 0);
}

/* checkMenuPath *************************************************************
 * if the file name of an application node is an absolute path name, this 
 * function checks if the first part of the path name corresponds to the value
 * of MENUPATH or HOME in the users environment. This is important if menu 
 * sets are copied from another user and the user forgets to adjust the file 
 * name for this menu set.
 ****************************************************************************/

checkMenuPath(aptr)
     ApplicationNode *aptr;
{
  char *getenv();
  char *menupath, *ptr;
  char fullName[256]; 

  if (!aptr)
    return(-1);
  
  if (!(menupath = getenv("MENUPATH"))) {    
    if (menupath = getenv("HOME"))
      exAlert(rme->menuEditDescr->meShell, -1, -1, 
	      "Environment variable MENUPATH\n is not specified;\n menu set is stored\nunder your home directory");;
  }
  
  if ((*aptr->file == '/') && 
      (strncmp(aptr->file, menupath, strlen(menupath)) != 0)) {
                                                /* wrong absolute path name */
    ptr = (char *) Malloc(strlen(aptr->applicationName) + 
			  strlen(MDS_FILESUFFIX));
    strcpy(ptr, aptr->applicationName);
    strcat(ptr, MDS_FILESUFFIX);
    if (makeFullName(menupath, ptr, fullName, 256))
      return(-1);

    setFile(aptr, fullName);
    free(ptr);
  }                                         

  return(1);
}

/* checkMenuAccess ***********************************************************
 * check if menu has read or write access
 ****************************************************************************/

Boolean checkMenuAccess(mptr)
     MHmenu *mptr;
{
  return((mptr->m_flags & (MDS_READACC | MDS_CHNGACC)) != 0);
}

/* delMenuUserFunct **********************************************************
 * function to delete user defined request functions (e.g. when menu is 
 * copied to another application node)
 ****************************************************************************/

Boolean delMenuUserFunct(mptr)
MHmenu *mptr;
{
  register i, j;
  Boolean FunctDef = False;

  if (mptr == (MHmenu *) 0) {
    setError(ERR_NULL);
    return (FunctDef);
  }

  for (i = 0; i < mptr->itemCnt; i++)
    for (j = 0; j < mptr->i_table[i]->i_return->reqCnt; j++)
      if (mptr->i_table[i]->i_requests[j]->flags == 1) {
	mptr->i_table[i]->i_requests[j]->flags = 0;
	mptr->i_table[i]->i_requests[j]->reqFnct = (int (*)()) 0;
	mptr->i_table[i]->i_requests[j]->reqData = (char *) 0;
	FunctDef = True;
      }

  return (FunctDef);
}

/* resetSelection **********************************************************
 * resets default items in menu description and current selections in the
 * corresponding runtime structure
 ****************************************************************************/

void resetSelection(mptr, menu)
     MHmenu *mptr;
     exMenu *menu;
{
  register i;

  for (i = 0; i < mptr->itemCnt; i++)
    resetDefaultItem(mptr->i_table[i]);

  for (i = 0; i < menu->itemCnt; i++)
    menu->selections[i] = 0;
  menu->selItemCnt = 0;
}

/* adjustDefaultAppl *********************************************************
 * searches for item in list menu corresponding to current application node
 ****************************************************************************/

adjustDefaultAppl(aptr)
     ApplicationNode *aptr;
{
  register i;
  MHmenu *alPtr,                  /* pointer to menu APPLLISTMENU */
         *adPtr;                  /* pointer to menu APPLDELMENU */

  if (!aptr)
    return(-1);

  setWorkApplication();

  if (!(alPtr = getMenuDescrByName(APPLLISTMENU)) || 
      !(adPtr = getMenuDescrByName(APPLDELMENU))) {
    setError(ERR_NULL);
    return(-1);
  }        /* adjust default item in list menus (e.g. after deletion) */

  for (i = 0; i < alPtr->itemCnt; i++) {
    resetDefaultItem(alPtr->i_table[i]);
    resetDefaultItem(adPtr->i_table[i]);
  }

  for (i = 0; i < alPtr->itemCnt; i++)  
    if (strcmp(aptr->applicationName, alPtr->i_table[i]->i_repr->name) == 0) {
      setDefaultItem(alPtr->i_table[i]);
      setDefaultItem(adPtr->i_table[i]);
      break;
    }

  return(0);
}

/* adjustDefaultMenu *********************************************************
 * searches for item in list menu corresponding to current menu
 ****************************************************************************/

adjustDefaultMenu(mptr)
     MHmenu *mptr;
{
  register i;
  MHmenu *mlPtr,                  /* pointer to menu MENULISTMENU */
         *mdPtr,                  /* pointer to menu MENUDELMENU */
         *mcPtr;                  /* pointer to menu MENUCOPY */

  if (!mptr)
    return(-1);

  setWorkApplication();

  if (!(mlPtr = getMenuDescrByName(MENULISTMENU)) || 
      !(mdPtr = getMenuDescrByName(MENUDELMENU)) || 
      !(mcPtr = getMenuDescrByName(MENUDELMENU))) {
    setError(ERR_NULL);
    return(-1);
  }        /* adjust default item in list menus (e.g. after deletion) */

  for (i = 0; i < mlPtr->itemCnt; i++) {
    resetDefaultItem(mlPtr->i_table[i]);
    resetDefaultItem(mdPtr->i_table[i]);
    resetDefaultItem(mcPtr->i_table[i]);
  }

  for (i = 0; i < mlPtr->itemCnt; i++)  
    if (strcmp(mptr->m_repr->name, mlPtr->i_table[i]->i_repr->name) == 0) {
      setDefaultItem(mlPtr->i_table[i]);
      setDefaultItem(mdPtr->i_table[i]);
      setDefaultItem(mcPtr->i_table[i]);
      break;
    }

  return(0);
}

/* meSetInitialValue *******************************************************
 * sets initial value to item of menu-description
 ****************************************************************************/

meSetInitialValue(mh, mds, item, req, initval)
     MHmenu *mh;
     exMenu *mds;
     int item, req;
     char *initval;     
{
  setRequestInitialValue(mh->i_table[item]->i_requests[req], initval);

  mds->entries[item]->flags |= MH_INITVALUE;
  mds->entries[item]->requests[req]->initialValue = strCopy(initval);
  mds->entries[item]->requests[req]->prelimInitialValue = strCopy(initval);
}

/****************************************************************************
 *  functions to give appropriate messages to the user
 ****************************************************************************/

/* checkApplAccess **********************************************************
 * notifies user if he does not have change access for application node
 ****************************************************************************/

checkApplAccess(aptr)
ApplicationNode *aptr;
{
  sprintf(msg, "You are not allowed to change\n attributes of application <%s>\n because you have no change access\n for application node <%s>!", 
	  aptr->applicationName, aptr->applicationName);
      
  exAlert(rme->menuEditDescr->meShell, -1, -1, msg);
}

/* ReplaceNotifier **********************************************************
 * asks user if a menu with the same name shall be overwritten 
 ****************************************************************************/

Boolean ReplaceNotifier(aptr, mptr)
     ApplicationNode *aptr;
     MHmenu *mptr;
{
  sprintf(msg, "Menu with name <%s>\n already exists under application node <%s>;\n Do you want to replace it?", mptr->m_repr->name, aptr->applicationName);

  return(exNotifyUser(rme->menuEditDescr->meShell, -1, -1, msg) == ANSWERYES);
}

/* SaveChangesNotifier *******************************************************
 * asks user if changes in current menu should be saved
 ****************************************************************************/

SaveChangesNotifier(aptr, mptr)
     ApplicationNode *aptr;
     MHmenu *mptr;
{
  sprintf(msg, "You changed menu <%s>;\n do you want to save it in\n the original menu set <%s>?", mptr->m_repr->name, aptr->applicationName);
      
  return(exNotify2User(rme->menuEditDescr->meShell, -1, -1, msg));
}

/* SaveNewMenuNotifier ******************************************************
 * asks user if menu created from him should be saved
 ****************************************************************************/

SaveNewMenuNotifier(aptr)
     ApplicationNode *aptr;
{
  sprintf(msg, "You created a new menu;\n do you want to save it in\n the current menu set <%s>?", aptr->applicationName);
      
  return(exNotify2User(rme->menuEditDescr->meShell, -1, -1, msg));
}

/* UserFunctNotifier ******************************************************
 * informs user that user functions are deleted when copying
 ****************************************************************************/

UserFunctNotifier(aptr, mptr)
     ApplicationNode *aptr;
     MHmenu *mptr;
{
  sprintf(msg, "Requests which include user defined request functions cannot be copied over to different application nodes; you have to redefine these requests for menu <%s> under application node <%s>!", 
	  mptr->m_repr->name, aptr->applicationName);

  exAlert(rme->menuEditDescr->meShell, -1, -1, msg);
}
